// ############################################################################################
//  vdspsb.h
//  Definition & implementation of _CVortexSoundBuffer class using direct access to /dev/dsp
//  -------------------------------------------------------------------------------------------
//  Copyright (C) 2004 Aloysius Indrayanto
//  This file is part of the VortexGE
//
//  This program is free software; you can redistribute it and/or modify it under the terms of 
//  the GNU Lesser General Public License as published by the Free Software Foundation; either
//  version 2.1 of the License, or (at your option) any later version.
//
//  This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
//  without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
//  See the GNU Lesser General Public License for more details.
//
//  You should have received a copy of the GNU Lesser General Public License along with this 
//  library; if not, write to the Free Software Foundation, Inc., 59 Temple Place, Suite 330, 
//  Boston, MA 02111-1307 USA.
// ############################################################################################
#ifndef _VDSPSB_H_
#define _VDSPSB_H_

// ============================================================================================
// Header files
// ============================================================================================
// ***** VortexGE headers
#include <vllsbcf.h>
#include <auddsp/vdspenv.h>

// ============================================================================================
// Reference to External variables
// ============================================================================================
// ***** The one and the only audio environment instance
extern _CVortexAudioEnv *_pcsAudioEnv_;

// ============================================================================================
// Exported classes definition
// ============================================================================================
// ***** The audio sound-buffer class
class _CVortexSoundBuffer
{
  protected:
    // Variables
    _TVortexSBData      mt_tD;
    _TVortexAudioFormat mt_tF;
    _vvoid (*_pfnGD_)(_TVortexPlayedSBData &tData, _vuchar *pucDstBuff);
    // Sound buffer functions
    inline _vvoid _GetData_(_TVortexPlayedSBData &tData, _vuchar *pucDstBuff);
    _vvoid _ExtractFormat_(_TVortexAudioFormat *ptFormat, _EVortexRawFormat eFormat);
    _vvoid _SwapEndian_(_vvoid);
    _vvoid _InvertSign_(_vvoid);
    _vvoid _Stereo2Mono_(_vvoid);
    _vbool _1628_(_vvoid);
    _vvoid _AssignCnvFunc_(_vbool bDifFreq);
    _vvoid _FreeBuffer_(_vvoid);

  public:
    // Constructor & destructor
    _CVortexSoundBuffer();
    ~_CVortexSoundBuffer();
    // Sound buffer functions
    inline _vbool Play(_vbool bLooping = false);
    _vbool AdjustAmplitude(_vfloat fLeftGain, _vfloat fRightGain);
    _vbool ConvertStereoToMono(_vvoid);
    _vbool Convert16BitTo8Bit(_vvoid);
    // File I/O functions
    _vbool LoadRAW(const _vchar *pcFileName, _EVortexRawFormat eFormat, _vint32 n32Frequency);
    _vbool LoadWAV(const _vchar *pcFileName);
    _vbool LoadVSW(const _vchar *pcFileName);
    _vbool SaveVSW(const _vchar *pcFileName);
  
  // Friend classes
  friend class _CVortexAudioEnv;
    
  // Friend functions
  friend _vvoid _DSPMixerU8m_(_vvoid);
  friend _vvoid _DSPMixerU8s_(_vvoid);
  friend _vvoid _DSPMixerS8m_(_vvoid);
  friend _vvoid _DSPMixerS8s_(_vvoid);
  friend _vvoid _DSPMixerU16m_(_vvoid);
  friend _vvoid _DSPMixerU16s_(_vvoid);
  friend _vvoid _DSPMixerS16m_(_vvoid);
  friend _vvoid _DSPMixerS16s_(_vvoid);
  friend _vvoid _DSPIMixerU16m_(_vvoid);
  friend _vvoid _DSPIMixerU16s_(_vvoid);
  friend _vvoid _DSPIMixerS16m_(_vvoid);
  friend _vvoid _DSPIMixerS16s_(_vvoid);
};

// ############################################################################################
// Inline functions implementation
// ############################################################################################
// ============================================================================================
// From exported classes
// ============================================================================================
// --------------------------------------------------------------------------------------------
// The audio sound-buffer class
// --------------------------------------------------------------------------------------------
// ***** Sound buffer functions
_vbool _CVortexSoundBuffer::Play(_vbool bLooping)
{
  // ..... Check
  if(_pcsAudioEnv_ == NULL) return(false);
  if(_pcsAudioEnv_->mt_tD.n32DSPFD < 0) return(false);
  if(mt_tD.pucLeft == NULL) return(false);
  
  // ..... Play sound buffer
  for(_vuint8 un8Cnt = 0; un8Cnt < _MAX_SB_; un8Cnt++)
  {
    if(_pcsAudioEnv_->mt_atPSB[un8Cnt].pcsSB == NULL)
    {
      _pcsAudioEnv_->mt_atPSB[un8Cnt].tData.pucLeft     = mt_tD.pucLeft;
      _pcsAudioEnv_->mt_atPSB[un8Cnt].tData.pucRight    = mt_tD.pucRight;
      _pcsAudioEnv_->mt_atPSB[un8Cnt].tData.pucLeftMax  = mt_tD.pucLeftMax;
      _pcsAudioEnv_->mt_atPSB[un8Cnt].tData.pucRightMax = mt_tD.pucRightMax;
      _pcsAudioEnv_->mt_atPSB[un8Cnt].tData.bPlaying    = true;
      _pcsAudioEnv_->mt_atPSB[un8Cnt].tData.bLooping    = bLooping;
      _pcsAudioEnv_->mt_atPSB[un8Cnt].tData.pucLeftCur  = mt_tD.pucLeft;
      _pcsAudioEnv_->mt_atPSB[un8Cnt].tData.pucRightCur = mt_tD.pucRight;
      _pcsAudioEnv_->mt_atPSB[un8Cnt].tData.un32PosMax  = mt_tD.un32PosMax;
      _pcsAudioEnv_->mt_atPSB[un8Cnt].tData.un32PosStep = mt_tD.un32PosStep;
      _pcsAudioEnv_->mt_atPSB[un8Cnt].tData.un32PosCur  = 0;
      _pcsAudioEnv_->mt_atPSB[un8Cnt].pcsSB = this;
      return(true);
    }
  }
  return(false);
}
_vvoid _CVortexSoundBuffer::_GetData_(_TVortexPlayedSBData &tData, _vuchar *pucDstBuff)
{
  _pfnGD_(tData, pucDstBuff);
}

#endif
// ============================================================================================
// End of vdspsb.h
// ============================================================================================
