// ############################################################################################
//  vbmp.h
//  Definition of structures used by Windows Bitmap
//  -------------------------------------------------------------------------------------------
//  Copyright (C) 2004 Aloysius Indrayanto
//  This file is part of the VortexGE
//
//  The source codes in this file is taken from the MS SDK header file "wingdi.h" with some
//  modifications
//
//  This program is free software; you can redistribute it and/or modify it under the terms of 
//  the GNU Lesser General Public License as published by the Free Software Foundation; either
//  version 2.1 of the License, or (at your option) any later version.
//
//  This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
//  without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
//  See the GNU Lesser General Public License for more details.
//
//  You should have received a copy of the GNU Lesser General Public License along with this 
//  library; if not, write to the Free Software Foundation, Inc., 59 Temple Place, Suite 330, 
//  Boston, MA 02111-1307 USA.
// ############################################################################################
#ifndef _VBMP_H_
#define _VBMP_H_

// ============================================================================================
// Header files
// ============================================================================================
// ***** VortexGE headers
#include <vbtypes.h>

// ============================================================================================
// Defines
// ============================================================================================
// ***** Used by Windows Bitmap
#define BI_RGB        _vuint32(0)
#define BI_RLE8       _vuint32(1)
#define BI_RLE4       _vuint32(2)
#define BI_BITFIELDS  _vuint32(3)
#define BI_JPEG       _vuint32(4)
#define BI_PNG        _vuint32(5)

// ============================================================================================
// Exported structures definition
// ============================================================================================
// ***** Used by Windows Bitmap
struct _TRGBTRIPLE
{
  _vuint8 un8B, un8G, un8R;
} __attribute__((packed));
struct _TRGBQUAD
{
  _vuint8 un8B, un8G, un8R, un8Reserved;
} __attribute__((packed));
struct _TBITMAPFILEHEADER
{
  _vuchar  aucBM[2];          // Always contains string "BM"
  _vuint32 un32Size;          // File size in bytes
  _vuint16 un16Reserved1;     // Reserved, must be zero
  _vuint16 un16Reserved2;     // Reserved, must be zero
  _vuint32 un32OffBits;       // Offset (in bytes) from the start of this structure to the bitmap bits
} __attribute__((packed));
struct _TBITMAPCOREHEADER
{
  _vuint32 un32Size;          // Size of this structure in bytes
  _vuint16 un16Width;         // Width of the bitmap
  _vuint16 un16Height;        // Height of the bitmap
  _vuint16 un16Planes;        // Number of planes for the target devices, must be one
  _vuint16 un16BitCount;      // The bits-per-pixel, must be 1, 4, 8, or 24
} __attribute__((packed));
struct _TBITMAPINFOHEADER
{
  _vuint32 un32Size;          // Size of this structure in bytes
  _vint32  n32Width;          // Width of the bitmap
  _vint32  n32Height;         // Height of the bitmap
  _vuint16 un16Planes;        // Number of planes for the target devices, must be one
  _vuint16 un16BitCount;      // The bits-per-pixel, must be 0, 1, 4, 8, 16, 24, or 32
  _vuint32 un32Compression;   // Must be BI_RGB, BI_RLE8, BI_RLE4, BI_BITFIELDS, BI_JPEG, or BI_PNG
  _vuint32 un32SizeImage;     // Size of the image, can be zero for BI_RGB
  _vint32  un32XPelsPerMeter; // Horizontal resolution in pixels-per-meter
  _vint32  un32YPelsPerMeter; // Vertical resolution in pixels-per-meter
  _vuint32 un32ClrUsed;       // Number of color indexes in the color table used by the bitmap
  _vuint32 un32ClrImportant;  // Number of color indexes table required for displaying the bitmap
} __attribute__((packed));

#endif
// ============================================================================================
// End of vbmp.h
// ============================================================================================
