// ############################################################################################
//  vbtypes.h
//  Definition of simple data types
//  -------------------------------------------------------------------------------------------
//  Copyright (C) 2004 Aloysius Indrayanto
//  This file is part of the VortexGE
//
//  This program is free software; you can redistribute it and/or modify it under the terms of 
//  the GNU Lesser General Public License as published by the Free Software Foundation; either
//  version 2.1 of the License, or (at your option) any later version.
//
//  This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
//  without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
//  See the GNU Lesser General Public License for more details.
//
//  You should have received a copy of the GNU Lesser General Public License along with this 
//  library; if not, write to the Free Software Foundation, Inc., 59 Temple Place, Suite 330, 
//  Boston, MA 02111-1307 USA.
// ############################################################################################
#ifndef _VBTYPES_H_
#define _VBTYPES_H_

// ############################################################################################
// #################################  B U I L D  M A C R O S  #################################
// ############################################################################################
// ####### You may change these macros to meet your needs, see "readme.txt" for details #######
// ############################################################################################
#define __USE_X11_VIDEO__
#define __USE_DSP_AUDIO__
#define __USE_ASM__
#define __USE_SSE2__
#define __USE_ALIGNED_MEMOPS__ 
#define __USE_COLORED_MESSAGES__
// ############################################################################################
// ###########################  E N D  O F  B U I L D  M A C R O S  ###########################
// ############################################################################################

// ============================================================================================
// Header files
// ============================================================================================
// ***** C++ headers
#include <stdio.h>

// ============================================================================================
// Special defines, DON'T CHANGE THEM !!!
// ============================================================================================
// ***** Create a complete version number (MJ.MN.RV)
#define __VG_COMB_VER__(MJ, MN, RV) ((_vuint32(MJ) << 24) | (_vuint32(MN) << 16) | (_vuint32(RV) << 8))

// ***** Define the current version of VortexGE
#define __VG_MAJOR__    0
#define __VG_MINOR__    4
#define __VG_REVISION__ 7
#define __VG_VERSION__  __VG_COMB_VER__(__VG_MAJOR__, __VG_MINOR__, __VG_REVISION__)

// ***** Check for minimum version required
#define __VG_MIN_VER__(MJ, MN, RV) (__VG_VERSION__ >= __VG_COMB_VER__(MJ, MN, RV))

// ***** Define alignment macros (needed for declaring memory-aligned variables)
//       If SSE or SSE2 assembly optimizations are used, all variables that will be involved
//       in these optimization must be aligned to a 16-byte boundary.
#if defined(i386) && defined(__USE_ASM__) && (defined(__USE_SSE__) || defined (__USE_SSE2__))
#define _sse __attribute__((aligned(16)))
#else
#define _sse
#endif

// ***** Define alignment directive for some buffers allocation. These buffers are :
//       + Declared inside <vidx11/vx11.h> :
//         - The render image  _TRenderImage::pucImage
//       + Declared inside <vvidbase.h> :
//         - The render buffer _TVortexRenderBuffer::mt_tRB.pun8Buffer
//         - The Z buffer      _TVortexRenderBuffer::mt_tRB.pn32ZBuffer
//       + Declared inside <vimage.h> :
//         - Texture buffers          _CVortexTexture::mt_pun8Buffer
//         - Animated texture buffers _CVortexAnimatedTexture::mt_pun8BufferArray
//         - Image buffers            _CVortexImage::mt_pun8Buffer
//         - Animated image buffers   _CVortexAnimatedImage::mt_pun8BufferArray
//       For now, only SSE instruction inside <vfastmem.h> will access these buffers,
//       so, only use address alignment if they need it.
//       HOWEVER, in the future, with more assembly optimizations, this directive may
//       needs to be always defined when SSE or SSE2 is used.
#if defined(i386) && defined(__USE_ASM__) && (defined(__USE_SSE__) || defined (__USE_SSE2__)) && defined(__USE_ALIGNED_MEMOPS__)
#define _ALIGN_BUFF_MEM_
#endif

// ***** Define a shortcut macro for simplicity (usable in x86 only)
//       __USE_ASM__ & __USE_SSE2__ : _ASM_ASM_, _ASM_MMX_, _ASM_SSE_, _ASM_SSE2_
//       __USE_ASM__ & __USE_SSE__  : _ASM_ASM_, _ASM_MMX_, _ASM_SSE_
//       __USE_ASM__ & __USE_MMX__  : _ASM_ASM_, _ASM_MMX_
//       __USE_ASM__                : _ASM_ASM_
#if defined(i386) && defined(__USE_ASM__)
# define _ASM_ASM_
# if defined(__USE_SSE2__)
#  define _ASM_SSE2_
#  define _ASM_SSE_
#  define _ASM_MMX_
# elif defined(__USE_SSE__)
#  define _ASM_SSE_
#  define _ASM_MMX_
# elif defined(__USE_MMX__)
#  define _ASM_MMX_
# endif  
#endif

// ***** Define a macro that gives the name of the current function
#if defined(__cplusplus) && (__GNUC_PREREQ(2, 6) || __GNUC_PREREQ(2, 4))
# define _N_ __PRETTY_FUNCTION__
#else
# if defined(__STDC_VERSION__) && (__STDC_VERSION__ >= 199901L)
#  define _N_ __func__
# else
#  define _N_ ((const _vchar *) 0)
# endif
#endif

// ***** Redefine __FILE__ and __LINE__ for simplicity 
#define _F_ __FILE__
#define _L_ __LINE__

// ============================================================================================
// Typedefs
// ============================================================================================
// ***** Characters
typedef char               _vchar;
typedef unsigned char      _vuchar;

// ***** Integers
typedef signed char        _vint8;
typedef signed short       _vint16;
typedef signed int         _vint32;
typedef signed long long   _vint64;
typedef unsigned char      _vuint8;
typedef unsigned short     _vuint16;
typedef unsigned int       _vuint32;
typedef unsigned long long _vuint64;

// ***** Integers (architecture-dependent size)
//       - In IA32 (x86) these will be 32-bit length, but in IA64 these will be 64-bit length.
//       - NEVER USE THESE DATA TYPES, unless for some specific operations such as converting
//         pointers to integers.
typedef signed long        _vintad;
typedef unsigned long      _vuintad;

// ***** Floating-points
typedef float              _vfloat;
typedef double             _vdouble;

// ***** Others
typedef bool               _vbool;
typedef void               _vvoid;
typedef FILE               _TVortexFILE;

#endif
// ============================================================================================
// End of vbtypes.h
// ============================================================================================
