// ############################################################################################
//  vfastmem.cpp
//  Implementation of fast-memory-manipulation functions
//  -------------------------------------------------------------------------------------------
//  Copyright (C) 2004 Aloysius Indrayanto
//  This file is part of the VortexGE
//
//  This program is free software; you can redistribute it and/or modify it under the terms of 
//  the GNU Lesser General Public License as published by the Free Software Foundation; either
//  version 2.1 of the License, or (at your option) any later version.
//
//  This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
//  without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
//  See the GNU Lesser General Public License for more details.
//
//  You should have received a copy of the GNU Lesser General Public License along with this 
//  library; if not, write to the Free Software Foundation, Inc., 59 Temple Place, Suite 330, 
//  Boston, MA 02111-1307 USA.
// ############################################################################################

// ============================================================================================
// Header files
// ============================================================================================
// ***** C++ headers
#include <errno.h>
#include <stdlib.h>
#include <string.h>
// ***** VortexGE headers
#include <vmdisp.h>
#include <vfastmem.h>

// ============================================================================================
// Exported functions implementation
// ============================================================================================
// ***** Aligned memory allocation functions
_vvoid *amalloc(_vuint32 un32Size, _vuint32 un32Alignment)
{
#ifdef __USE_XOPEN2K
  // ..... Temporary variables
  _vvoid   *pvTmp;
  _vuint32 un32R;

  // ..... Make sure that the alignment >= sizeof(_vvoid *)
  if(un32Alignment < sizeof(_vvoid *)) un32Alignment = sizeof(_vvoid *);
 
  // ..... Allocate aligned memory 
  un32R = posix_memalign(&pvTmp, un32Alignment, un32Size);
  if(un32R)
  {
    if(un32R == EINVAL)
    {
      VSetErrorCode(VERR_InvalidParameter);
      VError(_F_, _L_, "The alignment must be a power of two and a multiple of %d (size of _vvoid *)\n", sizeof(_vvoid *));
    }
    if(un32R == ENOMEM)
    {
      VSetErrorCode(VERR_NotEnoughMemory);
      VError(_F_, _L_, "Not enough memory\n");
    }
    return(NULL);
  }
  memset(pvTmp, 0, un32Size);
  return(pvTmp);
#else
  // ..... Temporary variables
  _vuintad unadPtr, unadBuff;

  // ..... Allocate aligned memory 
  VInfo(_F_, _L_, "posix_memalign(void **, size_t, size_t) is not supported.\nUsing manual address alignment, may cause performance degradation ...");
  unadPtr = (_vuintad) malloc(un32Size + un32Alignment + sizeof(_vvoid *));
  if(!unadPtr)
  {
    VSetErrorCode(VERR_NotEnoughMemory);
    VError(_F_, _L_, "Not enough memory\n");
    return(NULL);
  }
  memset((_vvoid *) unadPtr, 0, un32Size);
  unadBuff = (unadPtr + un32Alignment + sizeof(_vvoid *)) & (~(_vuintad(un32Alignment - 1)));
  *((_vuintad *)(unadBuff - sizeof(_vvoid *))) = unadPtr;
  return((_vvoid *) unadBuff);
#endif
}
_vvoid *agetrealaddress(_vvoid *pvBuffer)
{
#ifdef __USE_XOPEN2K
  return(pvBuffer);
#else
  _vvoid **ppvRealAddrRec = (_vvoid **) (((_vchar *) pvBuffer) - sizeof(_vvoid *));
  return(*ppvRealAddrRec);
#endif
}
_vvoid afree(_vvoid *pvBuffer)
{
  free(agetrealaddress(pvBuffer));
}

// ============================================================================================
// End of vfastmem.cpp
// ============================================================================================
