// ############################################################################################
//  vfrndrs.cpp
//  Implementation of fogged rendering functions of _CVortexPanel3D class
//  -------------------------------------------------------------------------------------------
//  Copyright (C) 2004 Aloysius Indrayanto
//  This file is part of the VortexGE
//
//  The source codes in this file is adopted from the book :
//    Cutting-Edge 3D Game Programming with C++
//    John De Goes
//    Coriolis Group, Inc.
//    USA, 1996
//
//  This program is free software; you can redistribute it and/or modify it under the terms of 
//  the GNU Lesser General Public License as published by the Free Software Foundation; either
//  version 2.1 of the License, or (at your option) any later version.
//
//  This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
//  without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
//  See the GNU Lesser General Public License for more details.
//
//  You should have received a copy of the GNU Lesser General Public License along with this 
//  library; if not, write to the Free Software Foundation, Inc., 59 Temple Place, Suite 330, 
//  Boston, MA 02111-1307 USA.
// ############################################################################################

// ============================================================================================
// Header files
// ============================================================================================
// ***** VortexGE headers
#include <vfcline.h>
#include <vpanel3d.h>

// ============================================================================================
// Exported classes implementation
// ============================================================================================
// --------------------------------------------------------------------------------------------
// The 3d panel class
// --------------------------------------------------------------------------------------------
// ***** Fogged, untextured, flat shaded
_vvoid _CVortexPanel3D::FogUFRender(_vvoid)
{
  // ..... Temporary variables
  _CVortexFogUFCeilLine csLeftSeg, csRightSeg;
  _vint32               n32RightPos, n32LeftPos, n32NewRightPos, n32NewLeftPos;
  _vint32               n32Height,  n32Width, n32XStart, n32XEnd;
  _vint32               n32ZStep, n32Z, n8EdgeCount, *pn32Z;
  _vuint32              un32YIndex, *pun32Buff;  
  _vuint8               un8Cnt;
  _vint32               n32F, n32FStep, n32Top = 0;

  // ..... Search for lowest Y coordinate (top of polygon)
  for(un8Cnt = 1; un8Cnt < mt_tP.un8SPCount; un8Cnt++)
    if(mt_tP.tSPoint[un8Cnt].n32Y < mt_tP.tSPoint[n32Top].n32Y) n32Top = un8Cnt;
  n32RightPos = n32LeftPos = n32Top;
    
  // ..... Calculate the index to the buffer
  un32YIndex = mt_tP.tSPoint[n32Top].n32Y * _pcsAct3DEnv_->mt_tRB->un32Width;
  
  // ..... Loop for all polygon edges
  n8EdgeCount = mt_tP.un8SPCount;
  while(n8EdgeCount > 0)
  {
    // Check if the right side of the polygon needs clipping
    if(csRightSeg.GetHeight() <= 0)
    {
      n32NewRightPos = n32RightPos + 1;
      if(n32NewRightPos >= mt_tP.un8SPCount) n32NewRightPos = 0;
      csRightSeg.Init(mt_tP.tSPoint[n32RightPos], mt_tP.tSPoint[n32NewRightPos]);
      n32RightPos = n32NewRightPos;
      n8EdgeCount--;
     
      // Perform clip on the top of edge
      if(csRightSeg.GetY() < _pcsAct3DEnv_->mt_n32MinY)
      {
        csRightSeg.ClipTop(_pcsAct3DEnv_->mt_n32MinY);
        un32YIndex = _pcsAct3DEnv_->mt_n32MinY  * _pcsAct3DEnv_->mt_tRB->un32Width;
      }
    }

    // Check if the left side of the polygon needs clipping
    if(csLeftSeg.GetHeight() <= 0)
    {
      n32NewLeftPos = n32LeftPos - 1;
      if(n32NewLeftPos < 0) n32NewLeftPos = mt_tP.un8SPCount - 1;
      csLeftSeg.Init(mt_tP.tSPoint[n32LeftPos], mt_tP.tSPoint[n32NewLeftPos]);
      n32LeftPos = n32NewLeftPos;
      n8EdgeCount--;
     
      // Perform clip on the top of edge
      if(csLeftSeg.GetY() < _pcsAct3DEnv_->mt_n32MinY)
      {
        csLeftSeg.ClipTop(_pcsAct3DEnv_->mt_n32MinY);
        un32YIndex = _pcsAct3DEnv_->mt_n32MinY  * _pcsAct3DEnv_->mt_tRB->un32Width;
      }
    }
    
    // Subdivide polygon into trapezoid
    if(csLeftSeg.GetHeight() < csRightSeg.GetHeight())
    {
      n32Height = csLeftSeg.GetHeight();
      if((csLeftSeg.GetY() + n32Height) > _pcsAct3DEnv_->mt_n32MaxY)
      {
        n32Height = _pcsAct3DEnv_->mt_n32MaxY - csLeftSeg.GetY();
        n8EdgeCount = 0;
      }
    }
    else
    {
      n32Height = csRightSeg.GetHeight();
      if((csRightSeg.GetY() + n32Height) > _pcsAct3DEnv_->mt_n32MaxY)
      {
        n32Height = _pcsAct3DEnv_->mt_n32MaxY - csRightSeg.GetY();
        n8EdgeCount = 0;
      }
    }
    
    // Loop for the height of the trapezoid
    while(n32Height-- > 0)
    {
      // Calculate initial values
      n32XStart = csLeftSeg.GetX();
      n32XEnd   = csRightSeg.GetX();
      n32Width  = n32XEnd - n32XStart;

      // Render
      if(n32Width > 0)
      {
        n32Z     = csLeftSeg.GetZ();
        n32ZStep = (csRightSeg.GetZ() - n32Z) / n32Width;
        
        n32F     = csLeftSeg.GetF();
        n32FStep = (csRightSeg.GetF() - n32F) / n32Width;

        // Clip the scan-line
        _FogUFClipHLine_(n32XStart, n32XEnd, n32Z, n32ZStep, n32F, n32FStep);
        n32Width  = n32XEnd - n32XStart;
        pn32Z     = &_pcsAct3DEnv_->mt_tRB->pn32ZBuffer[un32YIndex + _vuint32(n32XStart)];
        pun32Buff = (_vuint32 *) &_pcsAct3DEnv_->mt_tRB->pun8Buffer[(un32YIndex + _vuint32(n32XStart)) << 2];
        
        // Loop for the witdh of the scan-line
        while(n32Width-- > 0)
        {
          if(*pn32Z < n32Z)
          {
            *pn32Z     = n32Z;
            *pun32Buff = _pun32RFogColorTable_[(n32F >> _ISTEP_PREC_ << 5) | mt_tP.tTotal.un8R] |
                         _pun32GFogColorTable_[(n32F >> _ISTEP_PREC_ << 5) | mt_tP.tTotal.un8G] |
                         _pun32BFogColorTable_[(n32F >> _ISTEP_PREC_ << 5) | mt_tP.tTotal.un8B];
          }
          n32Z += n32ZStep;
          n32F += n32FStep;

          pn32Z++;
          pun32Buff++;
        }
      }
      csRightSeg.Step();
      csLeftSeg.Step();
      un32YIndex += _pcsAct3DEnv_->mt_tRB->un32Width;
    }
  }
}

// ***** Fogged, untextured, gouroud shaded
_vvoid _CVortexPanel3D::FogUGRender(_vvoid)
{
  // ..... Temporary variables
  _CVortexFogUGCeilLine csLeftSeg, csRightSeg;
  _vint32               n32RightPos, n32LeftPos, n32NewRightPos, n32NewLeftPos;
  _vint32               n32Height,  n32Width, n32XStart, n32XEnd;
  _vint32               n32ZStep, n32Z, n8EdgeCount, *pn32Z;
  _vint32               n32RStep, n32R, n32GStep, n32G, n32BStep, n32B;
  _vuint32              un32YIndex, *pun32Buff;  
  _vuint8               un8Cnt;
  _vint32               n32F, n32FStep, n32Top = 0;

  // ..... Search for lowest Y coordinate (top of polygon)
  for(un8Cnt = 1; un8Cnt < mt_tP.un8SPCount; un8Cnt++)
    if(mt_tP.tSPoint[un8Cnt].n32Y < mt_tP.tSPoint[n32Top].n32Y) n32Top = un8Cnt;
  n32RightPos = n32LeftPos = n32Top;
    
  // ..... Calculate the index to the buffer
  un32YIndex = mt_tP.tSPoint[n32Top].n32Y * _pcsAct3DEnv_->mt_tRB->un32Width;
  
  // ..... Loop for all polygon edges
  n8EdgeCount = mt_tP.un8SPCount;
  while(n8EdgeCount > 0)
  {
    // Check if the right side of the polygon needs clipping
    if(csRightSeg.GetHeight() <= 0)
    {
      n32NewRightPos = n32RightPos + 1;
      if(n32NewRightPos >= mt_tP.un8SPCount) n32NewRightPos = 0;
      csRightSeg.Init(mt_tP.tSPoint[n32RightPos], mt_tP.tSPoint[n32NewRightPos]);
      n32RightPos = n32NewRightPos;
      n8EdgeCount--;
     
      // Perform clip on the top of edge
      if(csRightSeg.GetY() < _pcsAct3DEnv_->mt_n32MinY)
      {
        csRightSeg.ClipTop(_pcsAct3DEnv_->mt_n32MinY);
        un32YIndex = _pcsAct3DEnv_->mt_n32MinY  * _pcsAct3DEnv_->mt_tRB->un32Width;
      }
    }

    // Check if the left side of the polygon needs clipping
    if(csLeftSeg.GetHeight() <= 0)
    {
      n32NewLeftPos = n32LeftPos - 1;
      if(n32NewLeftPos < 0) n32NewLeftPos = mt_tP.un8SPCount - 1;
      csLeftSeg.Init(mt_tP.tSPoint[n32LeftPos], mt_tP.tSPoint[n32NewLeftPos]);
      n32LeftPos = n32NewLeftPos;
      n8EdgeCount--;
     
      // Perform clip on the top of edge
      if(csLeftSeg.GetY() < _pcsAct3DEnv_->mt_n32MinY)
      {
        csLeftSeg.ClipTop(_pcsAct3DEnv_->mt_n32MinY);
        un32YIndex = _pcsAct3DEnv_->mt_n32MinY  * _pcsAct3DEnv_->mt_tRB->un32Width;
      }
    }
    
    // Subdivide polygon into trapezoid
    if(csLeftSeg.GetHeight() < csRightSeg.GetHeight())
    {
      n32Height = csLeftSeg.GetHeight();
      if((csLeftSeg.GetY() + n32Height) > _pcsAct3DEnv_->mt_n32MaxY)
      {
        n32Height = _pcsAct3DEnv_->mt_n32MaxY - csLeftSeg.GetY();
        n8EdgeCount = 0;
      }
    }
    else
    {
      n32Height = csRightSeg.GetHeight();
      if((csRightSeg.GetY() + n32Height) > _pcsAct3DEnv_->mt_n32MaxY)
      {
        n32Height = _pcsAct3DEnv_->mt_n32MaxY - csRightSeg.GetY();
        n8EdgeCount = 0;
      }
    }
    
    // Loop for the height of the trapezoid
    while(n32Height-- > 0)
    {
      // Calculate initial values
      n32XStart = csLeftSeg.GetX();
      n32XEnd   = csRightSeg.GetX();
      n32Width  = n32XEnd - n32XStart;

      // Render
      if(n32Width > 0)
      {
        n32Z     = csLeftSeg.GetZ();
        n32ZStep = (csRightSeg.GetZ() - n32Z) / n32Width;
        n32R     = csLeftSeg.GetR();
        n32RStep = (csRightSeg.GetR() - n32R) / n32Width;
        n32G     = csLeftSeg.GetG();
        n32GStep = (csRightSeg.GetG() - n32G) / n32Width;
        n32B     = csLeftSeg.GetB();
        n32BStep = (csRightSeg.GetB() - n32B) / n32Width;

        n32F     = csLeftSeg.GetF();
        n32FStep = (csRightSeg.GetF() - n32F) / n32Width;

        // Clip the scan-line
        _FogUGClipHLine_(n32XStart, n32XEnd, n32Z, n32ZStep, n32R, n32RStep, n32G, n32GStep, n32B, n32BStep, n32F, n32FStep);
        n32Width  = n32XEnd - n32XStart;
        pn32Z     = &_pcsAct3DEnv_->mt_tRB->pn32ZBuffer[un32YIndex + _vuint32(n32XStart)];
        pun32Buff = (_vuint32 *) &_pcsAct3DEnv_->mt_tRB->pun8Buffer[(un32YIndex + _vuint32(n32XStart)) << 2];
        
        // Loop for the witdh of the scan-line
        while(n32Width-- > 0)
        {
          if(*pn32Z < n32Z)
          {
            *pn32Z = n32Z;
            *pun32Buff = _pun32RFogColorTable_[(n32F >> _ISTEP_PREC_ << 5) | (n32R >> _ISTEP_PREC_)] |
                         _pun32GFogColorTable_[(n32F >> _ISTEP_PREC_ << 5) | (n32G >> _ISTEP_PREC_)] |
                         _pun32BFogColorTable_[(n32F >> _ISTEP_PREC_ << 5) | (n32B >> _ISTEP_PREC_)];
          }
          n32R += n32RStep;
          n32G += n32GStep;
          n32B += n32BStep;

          n32Z += n32ZStep;
          n32F += n32FStep;

          pn32Z++;
          pun32Buff++;
        }
      }
      csRightSeg.Step();
      csLeftSeg.Step();
      un32YIndex += _pcsAct3DEnv_->mt_tRB->un32Width;
    }
  }
}

// ***** Fogged, colored, flat shaded
_vvoid _CVortexPanel3D::FogCFRender(_vuint32 &un32R, _vuint32 &un32G, _vuint32 &un32B)
{
  // ..... Temporary variables
  _CVortexFogUFCeilLine csLeftSeg, csRightSeg;
  _vint32               n32RightPos, n32LeftPos, n32NewRightPos, n32NewLeftPos;
  _vint32               n32Height,  n32Width, n32XStart, n32XEnd;
  _vint32               n32ZStep, n32Z, n8EdgeCount, *pn32Z;
  _vuint32              un32YIndex, *pun32Buff;  
  _vuint8               un8Cnt;
  _vint32               n32F, n32FStep, n32Top = 0;

  // ..... Search for lowest Y coordinate (top of polygon)
  for(un8Cnt = 1; un8Cnt < mt_tP.un8SPCount; un8Cnt++)
    if(mt_tP.tSPoint[un8Cnt].n32Y < mt_tP.tSPoint[n32Top].n32Y) n32Top = un8Cnt;
  n32RightPos = n32LeftPos = n32Top;
    
  // ..... Calculate the index to the buffer
  un32YIndex = mt_tP.tSPoint[n32Top].n32Y * _pcsAct3DEnv_->mt_tRB->un32Width;
  
  // ..... Loop for all polygon edges
  n8EdgeCount = mt_tP.un8SPCount;
  while(n8EdgeCount > 0)
  {
    // Check if the right side of the polygon needs clipping
    if(csRightSeg.GetHeight() <= 0)
    {
      n32NewRightPos = n32RightPos + 1;
      if(n32NewRightPos >= mt_tP.un8SPCount) n32NewRightPos = 0;
      csRightSeg.Init(mt_tP.tSPoint[n32RightPos], mt_tP.tSPoint[n32NewRightPos]);
      n32RightPos = n32NewRightPos;
      n8EdgeCount--;
     
      // Perform clip on the top of edge
      if(csRightSeg.GetY() < _pcsAct3DEnv_->mt_n32MinY)
      {
        csRightSeg.ClipTop(_pcsAct3DEnv_->mt_n32MinY);
        un32YIndex = _pcsAct3DEnv_->mt_n32MinY  * _pcsAct3DEnv_->mt_tRB->un32Width;
      }
    }

    // Check if the left side of the polygon needs clipping
    if(csLeftSeg.GetHeight() <= 0)
    {
      n32NewLeftPos = n32LeftPos - 1;
      if(n32NewLeftPos < 0) n32NewLeftPos = mt_tP.un8SPCount - 1;
      csLeftSeg.Init(mt_tP.tSPoint[n32LeftPos], mt_tP.tSPoint[n32NewLeftPos]);
      n32LeftPos = n32NewLeftPos;
      n8EdgeCount--;
     
      // Perform clip on the top of edge
      if(csLeftSeg.GetY() < _pcsAct3DEnv_->mt_n32MinY)
      {
        csLeftSeg.ClipTop(_pcsAct3DEnv_->mt_n32MinY);
        un32YIndex = _pcsAct3DEnv_->mt_n32MinY  * _pcsAct3DEnv_->mt_tRB->un32Width;
      }
    }
    
    // Subdivide polygon into trapezoid
    if(csLeftSeg.GetHeight() < csRightSeg.GetHeight())
    {
      n32Height = csLeftSeg.GetHeight();
      if((csLeftSeg.GetY() + n32Height) > _pcsAct3DEnv_->mt_n32MaxY)
      {
        n32Height = _pcsAct3DEnv_->mt_n32MaxY - csLeftSeg.GetY();
        n8EdgeCount = 0;
      }
    }
    else
    {
      n32Height = csRightSeg.GetHeight();
      if((csRightSeg.GetY() + n32Height) > _pcsAct3DEnv_->mt_n32MaxY)
      {
        n32Height = _pcsAct3DEnv_->mt_n32MaxY - csRightSeg.GetY();
        n8EdgeCount = 0;
      }
    }
    
    // Loop for the height of the trapezoid
    while(n32Height-- > 0)
    {
      // Calculate initial values
      n32XStart = csLeftSeg.GetX();
      n32XEnd   = csRightSeg.GetX();
      n32Width  = n32XEnd - n32XStart;

      // Render
      if(n32Width > 0)
      {
        n32Z     = csLeftSeg.GetZ();
        n32ZStep = (csRightSeg.GetZ() - n32Z) / n32Width;

        n32F     = csLeftSeg.GetF();
        n32FStep = (csRightSeg.GetF() - n32F) / n32Width;
        
        // Clip the scan-line
        _FogUFClipHLine_(n32XStart, n32XEnd, n32Z, n32ZStep, n32F, n32FStep);
        n32Width  = n32XEnd - n32XStart;
        pn32Z     = &_pcsAct3DEnv_->mt_tRB->pn32ZBuffer[un32YIndex + _vuint32(n32XStart)];
        pun32Buff = (_vuint32 *) &_pcsAct3DEnv_->mt_tRB->pun8Buffer[(un32YIndex + _vuint32(n32XStart)) << 2];
        
        // Loop for the witdh of the scan-line
        while(n32Width-- > 0)
        {
          if(*pn32Z < n32Z)
          {
            *pn32Z     = n32Z;
            *pun32Buff = _pun32RFogLightTable_[(n32F >> _ISTEP_PREC_ << 10) | (un32R << 5) | mt_tP.tTotal.un8R] |
                         _pun32GFogLightTable_[(n32F >> _ISTEP_PREC_ << 10) | (un32G << 5) | mt_tP.tTotal.un8R] |
                         _pun32BFogLightTable_[(n32F >> _ISTEP_PREC_ << 10) | (un32B << 5) | mt_tP.tTotal.un8B];
          }
          n32Z += n32ZStep;
          n32F += n32FStep;

          pn32Z++;
          pun32Buff++;
        }
      }
      csRightSeg.Step();
      csLeftSeg.Step();
      un32YIndex += _pcsAct3DEnv_->mt_tRB->un32Width;
    }
  }
}

// ***** Fogged, colored, gouroud shaded
_vvoid _CVortexPanel3D::FogCGRender(_vuint32 &un32R, _vuint32 &un32G, _vuint32 &un32B)
{
  // ..... Temporary variables
  _CVortexFogUGCeilLine csLeftSeg, csRightSeg;
  _vint32               n32RightPos, n32LeftPos, n32NewRightPos, n32NewLeftPos;
  _vint32               n32Height,  n32Width, n32XStart, n32XEnd;
  _vint32               n32ZStep, n32Z, n8EdgeCount, *pn32Z;
  _vint32               n32RStep, n32R, n32GStep, n32G, n32BStep, n32B;
  _vuint32              un32YIndex, *pun32Buff;  
  _vuint8               un8Cnt;
  _vint32               n32F, n32FStep, n32Top = 0;

  // ..... Search for lowest Y coordinate (top of polygon)
  for(un8Cnt = 1; un8Cnt < mt_tP.un8SPCount; un8Cnt++)
    if(mt_tP.tSPoint[un8Cnt].n32Y < mt_tP.tSPoint[n32Top].n32Y) n32Top = un8Cnt;
  n32RightPos = n32LeftPos = n32Top;
    
  // ..... Calculate the index to the buffer
  un32YIndex = mt_tP.tSPoint[n32Top].n32Y * _pcsAct3DEnv_->mt_tRB->un32Width;
  
  // ..... Loop for all polygon edges
  n8EdgeCount = mt_tP.un8SPCount;
  while(n8EdgeCount > 0)
  {
    // Check if the right side of the polygon needs clipping
    if(csRightSeg.GetHeight() <= 0)
    {
      n32NewRightPos = n32RightPos + 1;
      if(n32NewRightPos >= mt_tP.un8SPCount) n32NewRightPos = 0;
      csRightSeg.Init(mt_tP.tSPoint[n32RightPos], mt_tP.tSPoint[n32NewRightPos]);
      n32RightPos = n32NewRightPos;
      n8EdgeCount--;
     
      // Perform clip on the top of edge
      if(csRightSeg.GetY() < _pcsAct3DEnv_->mt_n32MinY)
      {
        csRightSeg.ClipTop(_pcsAct3DEnv_->mt_n32MinY);
        un32YIndex = _pcsAct3DEnv_->mt_n32MinY  * _pcsAct3DEnv_->mt_tRB->un32Width;
      }
    }

    // Check if the left side of the polygon needs clipping
    if(csLeftSeg.GetHeight() <= 0)
    {
      n32NewLeftPos = n32LeftPos - 1;
      if(n32NewLeftPos < 0) n32NewLeftPos = mt_tP.un8SPCount - 1;
      csLeftSeg.Init(mt_tP.tSPoint[n32LeftPos], mt_tP.tSPoint[n32NewLeftPos]);
      n32LeftPos = n32NewLeftPos;
      n8EdgeCount--;
     
      // Perform clip on the top of edge
      if(csLeftSeg.GetY() < _pcsAct3DEnv_->mt_n32MinY)
      {
        csLeftSeg.ClipTop(_pcsAct3DEnv_->mt_n32MinY);
        un32YIndex = _pcsAct3DEnv_->mt_n32MinY  * _pcsAct3DEnv_->mt_tRB->un32Width;
      }
    }
    
    // Subdivide polygon into trapezoid
    if(csLeftSeg.GetHeight() < csRightSeg.GetHeight())
    {
      n32Height = csLeftSeg.GetHeight();
      if((csLeftSeg.GetY() + n32Height) > _pcsAct3DEnv_->mt_n32MaxY)
      {
        n32Height = _pcsAct3DEnv_->mt_n32MaxY - csLeftSeg.GetY();
        n8EdgeCount = 0;
      }
    }
    else
    {
      n32Height = csRightSeg.GetHeight();
      if((csRightSeg.GetY() + n32Height) > _pcsAct3DEnv_->mt_n32MaxY)
      {
        n32Height = _pcsAct3DEnv_->mt_n32MaxY - csRightSeg.GetY();
        n8EdgeCount = 0;
      }
    }
    
    // Loop for the height of the trapezoid
    while(n32Height-- > 0)
    {
      // Calculate initial values
      n32XStart = csLeftSeg.GetX();
      n32XEnd   = csRightSeg.GetX();
      n32Width  = n32XEnd - n32XStart;

      // Render
      if(n32Width > 0)
      {
        n32Z     = csLeftSeg.GetZ();
        n32ZStep = (csRightSeg.GetZ() - n32Z) / n32Width;
        n32R     = csLeftSeg.GetR();
        n32RStep = (csRightSeg.GetR() - n32R) / n32Width;
        n32G     = csLeftSeg.GetG();
        n32GStep = (csRightSeg.GetG() - n32G) / n32Width;
        n32B     = csLeftSeg.GetB();
        n32BStep = (csRightSeg.GetB() - n32B) / n32Width;

        n32F     = csLeftSeg.GetF();
        n32FStep = (csRightSeg.GetF() - n32F) / n32Width;

        // Clip the scan-line
        _FogUGClipHLine_(n32XStart, n32XEnd, n32Z, n32ZStep, n32R, n32RStep, n32G, n32GStep, n32B, n32BStep, n32F, n32FStep);
        n32Width  = n32XEnd - n32XStart;
        pn32Z     = &_pcsAct3DEnv_->mt_tRB->pn32ZBuffer[un32YIndex + _vuint32(n32XStart)];
        pun32Buff = (_vuint32 *) &_pcsAct3DEnv_->mt_tRB->pun8Buffer[(un32YIndex + _vuint32(n32XStart)) << 2];
        
        // Loop for the witdh of the scan-line
        while(n32Width-- > 0)
        {
          if(*pn32Z < n32Z)
          {
            *pn32Z = n32Z;
            *pun32Buff = _pun32RFogLightTable_[(n32F >> _ISTEP_PREC_ << 10) | (n32R >> _ISTEP_PREC_ << 5) | un32R] |
                         _pun32GFogLightTable_[(n32F >> _ISTEP_PREC_ << 10) | (n32G >> _ISTEP_PREC_ << 5) | un32G] |
                         _pun32BFogLightTable_[(n32F >> _ISTEP_PREC_ << 10) | (n32B >> _ISTEP_PREC_ << 5) | un32B];
          }
          n32R += n32RStep;
          n32G += n32GStep;
          n32B += n32BStep;

          n32Z += n32ZStep;
          n32F += n32FStep;

          pn32Z++;
          pun32Buff++;
        }
      }
      csRightSeg.Step();
      csLeftSeg.Step();
      un32YIndex += _pcsAct3DEnv_->mt_tRB->un32Width;
    }
  }
}

// ***** Fogged, textured, not shaded
_vvoid _CVortexPanel3D::FogTNRender(_CVortexTexture *pcsTexture)
{
  // ..... Temporary variables
  _CVortexFogTFCeilLine csLeftSeg, csRightSeg;
  _vint32               n32RightPos, n32LeftPos, n32NewRightPos, n32NewLeftPos;
  _vint32               n32Height,  n32Width, n32XStart, n32XEnd;
  _vint32               n32ZStep, n32Z, n8EdgeCount, *pn32Z;
  _vint32               n32UStep, n32U, n32VStep, n32V, n32Cnt, n32Len;
  _vint32               n32RZ, n32TX1, n32TY1, n32TX2, n32TY2, n32TXStep, n32TYStep, n32TX, n32TY;
  _vuint32              un32PreMultShift, un32Mask, un32YIndex, *pun32Buff;
  _vuint8               un8Cnt, *pun8TexBuff;
  _vint32               n32RunLoops, n32SubZStep = 0, n32SubUStep = 0, n32SubVStep = 0;
  _vint32               n32F, n32FStep, n32Pos, n32Top = 0;

  // ..... Prepare texturing
  pun8TexBuff      = pcsTexture->mt_pun8Buffer;
  un32PreMultShift = pcsTexture->mt_un32MultShift;
  un32Mask         = pcsTexture->mt_un32Mask;
  
  // ..... Search for lowest Y coordinate (top of polygon)
  for(un8Cnt = 1; un8Cnt < mt_tP.un8SPCount; un8Cnt++)
    if(mt_tP.tSPoint[un8Cnt].n32Y < mt_tP.tSPoint[n32Top].n32Y) n32Top = un8Cnt;
  n32RightPos = n32LeftPos = n32Top;
    
  // ..... Calculate the index to the buffer
  un32YIndex = mt_tP.tSPoint[n32Top].n32Y * _pcsAct3DEnv_->mt_tRB->un32Width;
  
  // ..... Loop for all polygon edges
  n8EdgeCount = mt_tP.un8SPCount;
  while(n8EdgeCount > 0)
  {
    // Check if the right side of the polygon needs clipping
    if(csRightSeg.GetHeight() <= 0)
    {
      n32NewRightPos = n32RightPos + 1;
      if(n32NewRightPos >= mt_tP.un8SPCount) n32NewRightPos = 0;
      csRightSeg.Init(mt_tP.tSPoint[n32RightPos], mt_tP.tSPoint[n32NewRightPos]);
      n32RightPos = n32NewRightPos;
      n8EdgeCount--;
     
      // Perform clip on the top of edge
      if(csRightSeg.GetY() < _pcsAct3DEnv_->mt_n32MinY)
      {
        csRightSeg.ClipTop(_pcsAct3DEnv_->mt_n32MinY);
        un32YIndex = _pcsAct3DEnv_->mt_n32MinY  * _pcsAct3DEnv_->mt_tRB->un32Width;
      }
    }

    // Check if the left side of the polygon needs clipping
    if(csLeftSeg.GetHeight() <= 0)
    {
      n32NewLeftPos = n32LeftPos - 1;
      if(n32NewLeftPos < 0) n32NewLeftPos = mt_tP.un8SPCount - 1;
      csLeftSeg.Init(mt_tP.tSPoint[n32LeftPos], mt_tP.tSPoint[n32NewLeftPos]);
      n32LeftPos = n32NewLeftPos;
      n8EdgeCount--;
     
      // Perform clip on the top of edge
      if(csLeftSeg.GetY() < _pcsAct3DEnv_->mt_n32MinY)
      {
        csLeftSeg.ClipTop(_pcsAct3DEnv_->mt_n32MinY);
        un32YIndex = _pcsAct3DEnv_->mt_n32MinY  * _pcsAct3DEnv_->mt_tRB->un32Width;
      }
    }
    
    // Subdivide polygon into trapezoid
    if(csLeftSeg.GetHeight() < csRightSeg.GetHeight())
    {
      n32Height = csLeftSeg.GetHeight();
      if((csLeftSeg.GetY() + n32Height) > _pcsAct3DEnv_->mt_n32MaxY)
      {
        n32Height = _pcsAct3DEnv_->mt_n32MaxY - csLeftSeg.GetY();
        n8EdgeCount = 0;
      }
    }
    else
    {
      n32Height = csRightSeg.GetHeight();
      if((csRightSeg.GetY() + n32Height) > _pcsAct3DEnv_->mt_n32MaxY)
      {
        n32Height = _pcsAct3DEnv_->mt_n32MaxY - csRightSeg.GetY();
        n8EdgeCount = 0;
      }
    }
    
    // Loop for the height of the trapezoid
    while(n32Height-- > 0)
    {
      // Calculate initial values
      n32XStart = csLeftSeg.GetX();
      n32XEnd   = csRightSeg.GetX();
      n32Width  = n32XEnd - n32XStart;

      // Render
      if(n32Width > 0)
      {
        n32Z     = csLeftSeg.GetZ();
        n32ZStep = (csRightSeg.GetZ() - n32Z) / n32Width;
        n32U     = csLeftSeg.GetU();
        n32UStep = (csRightSeg.GetU() - n32U) / n32Width;
        n32V     = csLeftSeg.GetV();
        n32VStep = (csRightSeg.GetV() - n32V) / n32Width;

        n32F     = csLeftSeg.GetF();
        n32FStep = (csRightSeg.GetF() - n32F) / n32Width;
        
        // Clip the scan-line
        _FogTFClipHLine_(n32XStart, n32XEnd, n32Z, n32ZStep, n32U, n32UStep, n32V, n32VStep, n32F, n32FStep);
        n32Width  = n32XEnd - n32XStart;
        pn32Z     = &_pcsAct3DEnv_->mt_tRB->pn32ZBuffer[un32YIndex + _vuint32(n32XStart)];
        pun32Buff = (_vuint32 *) &_pcsAct3DEnv_->mt_tRB->pun8Buffer[(un32YIndex + _vuint32(n32XStart)) << 2];
        
        // Calculate subdivision loops
        n32RunLoops = n32Width / _RUNLEN_;
        
        // Premultiply steps for subdivision
        if(n32Width > 0)
        {
          n32SubZStep = n32ZStep << _RUNSHIFT_;
          n32SubUStep = n32UStep << _RUNSHIFT_;
          n32SubVStep = n32VStep << _RUNSHIFT_;
        }
        
        // Loop for number of subdivisions
        for(n32Cnt = 0; n32Cnt < n32RunLoops; n32Cnt++)
        {
          n32Len = _RUNLEN_;

          n32RZ  = n32Z - _pcsAct3DEnv_->mt_tRB->n32ZTrans;
          n32TX1 = (n32U / n32RZ) << _LSTEP_PREC_;
          n32TY1 = (n32V / n32RZ) << _LSTEP_PREC_;

          n32RZ += n32SubZStep;
          n32U  += n32SubUStep;
          n32V  += n32SubVStep;

          n32TX2 = (n32U / n32RZ) << _LSTEP_PREC_;
          n32TY2 = (n32V / n32RZ) << _LSTEP_PREC_;
          
          n32TXStep = (n32TX2 - n32TX1) >> _RUNSHIFT_;
          n32TYStep = ((n32TY2 - n32TY1) >> _RUNSHIFT_) << un32PreMultShift;
          
          n32TX = n32TX1;
          n32TY = n32TY1 << un32PreMultShift;

          // Rasterize the scanline
          while(n32Len-- > 0)
          {
            if(*pn32Z < n32Z)
            {
              *pn32Z = n32Z;

              n32Pos = ((n32TY & un32Mask) + n32TX) >> _LSTEP_PREC_ << 2;
              *pun32Buff = _pun32RFogColorTable_[(n32F >> _ISTEP_PREC_ << 5) | pun8TexBuff[n32Pos + 2]] |
                           _pun32GFogColorTable_[(n32F >> _ISTEP_PREC_ << 5) | pun8TexBuff[n32Pos + 1]] |
                           _pun32BFogColorTable_[(n32F >> _ISTEP_PREC_ << 5) | pun8TexBuff[n32Pos + 0]];              
            }
            n32Z  += n32ZStep;
            n32TX += n32TXStep;
            n32TY += n32TYStep;

            n32F += n32FStep;
            
            pn32Z++;
            pun32Buff++; 
          }
        }
        
        // Rasterize the remainder of the scanline
        n32Len = n32Width & (_RUNLEN_ - 1);
        if(n32Len > 0)
        {
          n32RZ  = n32Z - _pcsAct3DEnv_->mt_tRB->n32ZTrans;
          n32TX1 = (n32U / n32RZ) << _LSTEP_PREC_;
          n32TY1 = (n32V / n32RZ) << _LSTEP_PREC_;

          n32RZ += (n32ZStep * n32Len);
          n32U  += (n32UStep * n32Len);
          n32V  += (n32VStep * n32Len);

          n32TX2 = (n32U / n32RZ) << _LSTEP_PREC_;
          n32TY2 = (n32V / n32RZ) << _LSTEP_PREC_;
          
          n32TXStep = (n32TX2 - n32TX1) / n32Len;
          n32TYStep = ((n32TY2 - n32TY1) / n32Len) << un32PreMultShift;

          n32TX = n32TX1;
          n32TY = n32TY1 << un32PreMultShift;

          while(n32Len-- > 0)
          {
            if(*pn32Z < n32Z)
            {
              *pn32Z = n32Z;
                           
              n32Pos = ((n32TY & un32Mask) + n32TX) >> _LSTEP_PREC_ << 2;
              *pun32Buff = _pun32RFogColorTable_[(n32F >> _ISTEP_PREC_ << 5) | pun8TexBuff[n32Pos + 2]] |
                           _pun32GFogColorTable_[(n32F >> _ISTEP_PREC_ << 5) | pun8TexBuff[n32Pos + 1]] |
                           _pun32BFogColorTable_[(n32F >> _ISTEP_PREC_ << 5) | pun8TexBuff[n32Pos + 0]];              
            }
            n32Z  += n32ZStep;
            n32TX += n32TXStep;
            n32TY += n32TYStep;

            n32F += n32FStep;

            pn32Z++;
            pun32Buff++; 
          }          
        }
      }
      csRightSeg.Step();
      csLeftSeg.Step();
      un32YIndex += _pcsAct3DEnv_->mt_tRB->un32Width;
    }
  }
}

// ***** Fogged, textured, flat shaded
_vvoid _CVortexPanel3D::FogTFRender(_CVortexTexture *pcsTexture)
{
  // ..... Temporary variables
  _CVortexFogTFCeilLine csLeftSeg, csRightSeg;
  _vint32               n32RightPos, n32LeftPos, n32NewRightPos, n32NewLeftPos;
  _vint32               n32Height,  n32Width, n32XStart, n32XEnd;
  _vint32               n32ZStep, n32Z, n8EdgeCount, *pn32Z;
  _vint32               n32UStep, n32U, n32VStep, n32V, n32Cnt, n32Len;
  _vint32               n32RZ, n32TX1, n32TY1, n32TX2, n32TY2, n32TXStep, n32TYStep, n32TX, n32TY;
  _vuint32              un32PreMultShift, un32Mask, un32YIndex, *pun32Buff;
  _vuint8               un8Cnt, *pun8TexBuff;
  _vint32               n32RunLoops, n32SubZStep = 0, n32SubUStep = 0, n32SubVStep = 0;
  _vint32               n32F, n32FStep, n32Pos, n32Top = 0;

  // ..... Prepare texturing
  pun8TexBuff      = pcsTexture->mt_pun8Buffer;
  un32PreMultShift = pcsTexture->mt_un32MultShift;
  un32Mask         = pcsTexture->mt_un32Mask;
  
  // ..... Search for lowest Y coordinate (top of polygon)
  for(un8Cnt = 1; un8Cnt < mt_tP.un8SPCount; un8Cnt++)
    if(mt_tP.tSPoint[un8Cnt].n32Y < mt_tP.tSPoint[n32Top].n32Y) n32Top = un8Cnt;
  n32RightPos = n32LeftPos = n32Top;
    
  // ..... Calculate the index to the buffer
  un32YIndex = mt_tP.tSPoint[n32Top].n32Y * _pcsAct3DEnv_->mt_tRB->un32Width;
  
  // ..... Loop for all polygon edges
  n8EdgeCount = mt_tP.un8SPCount;
  while(n8EdgeCount > 0)
  {
    // Check if the right side of the polygon needs clipping
    if(csRightSeg.GetHeight() <= 0)
    {
      n32NewRightPos = n32RightPos + 1;
      if(n32NewRightPos >= mt_tP.un8SPCount) n32NewRightPos = 0;
      csRightSeg.Init(mt_tP.tSPoint[n32RightPos], mt_tP.tSPoint[n32NewRightPos]);
      n32RightPos = n32NewRightPos;
      n8EdgeCount--;
     
      // Perform clip on the top of edge
      if(csRightSeg.GetY() < _pcsAct3DEnv_->mt_n32MinY)
      {
        csRightSeg.ClipTop(_pcsAct3DEnv_->mt_n32MinY);
        un32YIndex = _pcsAct3DEnv_->mt_n32MinY  * _pcsAct3DEnv_->mt_tRB->un32Width;
      }
    }

    // Check if the left side of the polygon needs clipping
    if(csLeftSeg.GetHeight() <= 0)
    {
      n32NewLeftPos = n32LeftPos - 1;
      if(n32NewLeftPos < 0) n32NewLeftPos = mt_tP.un8SPCount - 1;
      csLeftSeg.Init(mt_tP.tSPoint[n32LeftPos], mt_tP.tSPoint[n32NewLeftPos]);
      n32LeftPos = n32NewLeftPos;
      n8EdgeCount--;
     
      // Perform clip on the top of edge
      if(csLeftSeg.GetY() < _pcsAct3DEnv_->mt_n32MinY)
      {
        csLeftSeg.ClipTop(_pcsAct3DEnv_->mt_n32MinY);
        un32YIndex = _pcsAct3DEnv_->mt_n32MinY  * _pcsAct3DEnv_->mt_tRB->un32Width;
      }
    }
    
    // Subdivide polygon into trapezoid
    if(csLeftSeg.GetHeight() < csRightSeg.GetHeight())
    {
      n32Height = csLeftSeg.GetHeight();
      if((csLeftSeg.GetY() + n32Height) > _pcsAct3DEnv_->mt_n32MaxY)
      {
        n32Height = _pcsAct3DEnv_->mt_n32MaxY - csLeftSeg.GetY();
        n8EdgeCount = 0;
      }
    }
    else
    {
      n32Height = csRightSeg.GetHeight();
      if((csRightSeg.GetY() + n32Height) > _pcsAct3DEnv_->mt_n32MaxY)
      {
        n32Height = _pcsAct3DEnv_->mt_n32MaxY - csRightSeg.GetY();
        n8EdgeCount = 0;
      }
    }
    
    // Loop for the height of the trapezoid
    while(n32Height-- > 0)
    {
      // Calculate initial values
      n32XStart = csLeftSeg.GetX();
      n32XEnd   = csRightSeg.GetX();
      n32Width  = n32XEnd - n32XStart;

      // Render
      if(n32Width > 0)
      {
        n32Z     = csLeftSeg.GetZ();
        n32ZStep = (csRightSeg.GetZ() - n32Z) / n32Width;
        n32U     = csLeftSeg.GetU();
        n32UStep = (csRightSeg.GetU() - n32U) / n32Width;
        n32V     = csLeftSeg.GetV();
        n32VStep = (csRightSeg.GetV() - n32V) / n32Width;

        n32F     = csLeftSeg.GetF();
        n32FStep = (csRightSeg.GetF() - n32F) / n32Width;

        // Clip the scan-line
        _FogTFClipHLine_(n32XStart, n32XEnd, n32Z, n32ZStep, n32U, n32UStep, n32V, n32VStep, n32F, n32FStep);
        n32Width  = n32XEnd - n32XStart;
        pn32Z     = &_pcsAct3DEnv_->mt_tRB->pn32ZBuffer[un32YIndex + _vuint32(n32XStart)];
        pun32Buff = (_vuint32 *) &_pcsAct3DEnv_->mt_tRB->pun8Buffer[(un32YIndex + _vuint32(n32XStart)) << 2];
        
        // Calculate subdivision loops
        n32RunLoops = n32Width / _RUNLEN_;
        
        // Premultiply steps for subdivision
        if(n32Width > 0)
        {
          n32SubZStep = n32ZStep << _RUNSHIFT_;
          n32SubUStep = n32UStep << _RUNSHIFT_;
          n32SubVStep = n32VStep << _RUNSHIFT_;
        }
        
        // Loop for number of subdivisions
        for(n32Cnt = 0; n32Cnt < n32RunLoops; n32Cnt++)
        {
          n32Len = _RUNLEN_;

          n32RZ  = n32Z - _pcsAct3DEnv_->mt_tRB->n32ZTrans;
          n32TX1 = (n32U / n32RZ) << _LSTEP_PREC_;
          n32TY1 = (n32V / n32RZ) << _LSTEP_PREC_;

          n32RZ += n32SubZStep;
          n32U  += n32SubUStep;
          n32V  += n32SubVStep;

          n32TX2 = (n32U / n32RZ) << _LSTEP_PREC_;
          n32TY2 = (n32V / n32RZ) << _LSTEP_PREC_;
          
          n32TXStep = (n32TX2 - n32TX1) >> _RUNSHIFT_;
          n32TYStep = ((n32TY2 - n32TY1) >> _RUNSHIFT_) << un32PreMultShift;
          
          n32TX = n32TX1;
          n32TY = n32TY1 << un32PreMultShift;

          // Rasterize the scanline
          while(n32Len-- > 0)
          {
            if(*pn32Z < n32Z)
            {
              *pn32Z = n32Z;

              n32Pos = ((n32TY & un32Mask) + n32TX) >> _LSTEP_PREC_ << 2;
              *pun32Buff = _pun32RFogLightTable_[(n32F >> _ISTEP_PREC_ << 10) | (mt_tP.tTotal.un8R << 5) | pun8TexBuff[n32Pos + 2]] |
                           _pun32GFogLightTable_[(n32F >> _ISTEP_PREC_ << 10) | (mt_tP.tTotal.un8G << 5) | pun8TexBuff[n32Pos + 1]] |
                           _pun32BFogLightTable_[(n32F >> _ISTEP_PREC_ << 10) | (mt_tP.tTotal.un8B << 5) | pun8TexBuff[n32Pos + 0]];
            }
            n32Z  += n32ZStep;
            n32TX += n32TXStep;
            n32TY += n32TYStep;
            
            n32F += n32FStep;

            pn32Z++;
            pun32Buff++; 
          }
        }
        
        // Rasterize the remainder of the scanline
        n32Len = n32Width & (_RUNLEN_ - 1);
        if(n32Len > 0)
        {
          n32RZ  = n32Z - _pcsAct3DEnv_->mt_tRB->n32ZTrans;
          n32TX1 = (n32U / n32RZ) << _LSTEP_PREC_;
          n32TY1 = (n32V / n32RZ) << _LSTEP_PREC_;

          n32RZ += (n32ZStep * n32Len);
          n32U  += (n32UStep * n32Len);
          n32V  += (n32VStep * n32Len);

          n32TX2 = (n32U / n32RZ) << _LSTEP_PREC_;
          n32TY2 = (n32V / n32RZ) << _LSTEP_PREC_;
          
          n32TXStep = (n32TX2 - n32TX1) / n32Len;
          n32TYStep = ((n32TY2 - n32TY1) / n32Len) << un32PreMultShift;

          n32TX = n32TX1;
          n32TY = n32TY1 << un32PreMultShift;

          while(n32Len-- > 0)
          {
            if(*pn32Z < n32Z)
            {
              *pn32Z = n32Z;
                           
              n32Pos = ((n32TY & un32Mask) + n32TX) >> _LSTEP_PREC_ << 2;
              *pun32Buff = _pun32RFogLightTable_[(n32F >> _ISTEP_PREC_ << 10) | (mt_tP.tTotal.un8R << 5) | pun8TexBuff[n32Pos + 2]] |
                           _pun32GFogLightTable_[(n32F >> _ISTEP_PREC_ << 10) | (mt_tP.tTotal.un8G << 5) | pun8TexBuff[n32Pos + 1]] |
                           _pun32BFogLightTable_[(n32F >> _ISTEP_PREC_ << 10) | (mt_tP.tTotal.un8B << 5) | pun8TexBuff[n32Pos + 0]];
            }
            n32Z  += n32ZStep;
            n32TX += n32TXStep;
            n32TY += n32TYStep;

            n32F += n32FStep;

            pn32Z++;
            pun32Buff++; 
          }          
        }
      }
      csRightSeg.Step();
      csLeftSeg.Step();
      un32YIndex += _pcsAct3DEnv_->mt_tRB->un32Width;
    }
  }
}

// ***** Fogged, textured, gouroud shaded
_vvoid _CVortexPanel3D::FogTGRender(_CVortexTexture *pcsTexture)
{
  // ..... Temporary variables
  _CVortexFogTGCeilLine csLeftSeg, csRightSeg;
  _vint32               n32RightPos, n32LeftPos, n32NewRightPos, n32NewLeftPos;
  _vint32               n32Height,  n32Width, n32XStart, n32XEnd;
  _vint32               n32ZStep, n32Z, n8EdgeCount, *pn32Z;
  _vint32               n32UStep, n32U, n32VStep, n32V, n32Cnt, n32Len;
  _vint32               n32RZ, n32TX1, n32TY1, n32TX2, n32TY2, n32TXStep, n32TYStep, n32TX, n32TY;
  _vuint32              un32PreMultShift, un32Mask, un32YIndex, *pun32Buff;
  _vuint8               un8Cnt, *pun8TexBuff;
  _vint32               n32RStep, n32R, n32GStep, n32G, n32BStep, n32B, n32F, n32FStep;
  _vint32               n32Pos, n32Top = 0;
  _vint32               n32RunLoops, n32SubZStep = 0, n32SubUStep = 0, n32SubVStep = 0;

  // ..... Prepare texturing
  pun8TexBuff      = pcsTexture->mt_pun8Buffer;
  un32PreMultShift = pcsTexture->mt_un32MultShift;
  un32Mask         = pcsTexture->mt_un32Mask;
  
  // ..... Search for lowest Y coordinate (top of polygon)
  for(un8Cnt = 1; un8Cnt < mt_tP.un8SPCount; un8Cnt++)
    if(mt_tP.tSPoint[un8Cnt].n32Y < mt_tP.tSPoint[n32Top].n32Y) n32Top = un8Cnt;
  n32RightPos = n32LeftPos = n32Top;
    
  // ..... Calculate the index to the buffer
  un32YIndex = mt_tP.tSPoint[n32Top].n32Y * _pcsAct3DEnv_->mt_tRB->un32Width;
  
  // ..... Loop for all polygon edges
  n8EdgeCount = mt_tP.un8SPCount;
  while(n8EdgeCount > 0)
  {
    // Check if the right side of the polygon needs clipping
    if(csRightSeg.GetHeight() <= 0)
    {
      n32NewRightPos = n32RightPos + 1;
      if(n32NewRightPos >= mt_tP.un8SPCount) n32NewRightPos = 0;
      csRightSeg.Init(mt_tP.tSPoint[n32RightPos], mt_tP.tSPoint[n32NewRightPos]);
      n32RightPos = n32NewRightPos;
      n8EdgeCount--;
     
      // Perform clip on the top of edge
      if(csRightSeg.GetY() < _pcsAct3DEnv_->mt_n32MinY)
      {
        csRightSeg.ClipTop(_pcsAct3DEnv_->mt_n32MinY);
        un32YIndex = _pcsAct3DEnv_->mt_n32MinY  * _pcsAct3DEnv_->mt_tRB->un32Width;
      }
    }

    // Check if the left side of the polygon needs clipping
    if(csLeftSeg.GetHeight() <= 0)
    {
      n32NewLeftPos = n32LeftPos - 1;
      if(n32NewLeftPos < 0) n32NewLeftPos = mt_tP.un8SPCount - 1;
      csLeftSeg.Init(mt_tP.tSPoint[n32LeftPos], mt_tP.tSPoint[n32NewLeftPos]);
      n32LeftPos = n32NewLeftPos;
      n8EdgeCount--;
     
      // Perform clip on the top of edge
      if(csLeftSeg.GetY() < _pcsAct3DEnv_->mt_n32MinY)
      {
        csLeftSeg.ClipTop(_pcsAct3DEnv_->mt_n32MinY);
        un32YIndex = _pcsAct3DEnv_->mt_n32MinY  * _pcsAct3DEnv_->mt_tRB->un32Width;
      }
    }
    
    // Subdivide polygon into trapezoid
    if(csLeftSeg.GetHeight() < csRightSeg.GetHeight())
    {
      n32Height = csLeftSeg.GetHeight();
      if((csLeftSeg.GetY() + n32Height) > _pcsAct3DEnv_->mt_n32MaxY)
      {
        n32Height = _pcsAct3DEnv_->mt_n32MaxY - csLeftSeg.GetY();
        n8EdgeCount = 0;
      }
    }
    else
    {
      n32Height = csRightSeg.GetHeight();
      if((csRightSeg.GetY() + n32Height) > _pcsAct3DEnv_->mt_n32MaxY)
      {
        n32Height = _pcsAct3DEnv_->mt_n32MaxY - csRightSeg.GetY();
        n8EdgeCount = 0;
      }
    }
    
    // Loop for the height of the trapezoid
    while(n32Height-- > 0)
    {
      // Calculate initial values
      n32XStart = csLeftSeg.GetX();
      n32XEnd   = csRightSeg.GetX();
      n32Width  = n32XEnd - n32XStart;

      // Render
      if(n32Width > 0)
      {
        n32Z     = csLeftSeg.GetZ();
        n32ZStep = (csRightSeg.GetZ() - n32Z) / n32Width;
        n32U     = csLeftSeg.GetU();
        n32UStep = (csRightSeg.GetU() - n32U) / n32Width;
        n32V     = csLeftSeg.GetV();
        n32VStep = (csRightSeg.GetV() - n32V) / n32Width;

        n32R     = csLeftSeg.GetR();
        n32RStep = (csRightSeg.GetR() - n32R) / n32Width;
        n32G     = csLeftSeg.GetG();
        n32GStep = (csRightSeg.GetG() - n32G) / n32Width;
        n32B     = csLeftSeg.GetB();
        n32BStep = (csRightSeg.GetB() - n32B) / n32Width;
        
        n32F     = csLeftSeg.GetF();
        n32FStep = (csRightSeg.GetF() - n32F) / n32Width;

        // Clip the scan-line
        _FogTGClipHLine_(n32XStart, n32XEnd, n32Z, n32ZStep, n32U, n32UStep, n32V, n32VStep, n32R, n32RStep, n32G, n32GStep, n32B, n32BStep, n32F, n32FStep);
        n32Width  = n32XEnd - n32XStart;
        pn32Z     = &_pcsAct3DEnv_->mt_tRB->pn32ZBuffer[un32YIndex + _vuint32(n32XStart)];
        pun32Buff = (_vuint32 *) &_pcsAct3DEnv_->mt_tRB->pun8Buffer[(un32YIndex + _vuint32(n32XStart)) << 2];
        
        // Calculate subdivision loops
        n32RunLoops = n32Width / _RUNLEN_;
        
        // Premultiply steps for subdivision
        if(n32Width > 0)
        {
          n32SubZStep = n32ZStep << _RUNSHIFT_;
          n32SubUStep = n32UStep << _RUNSHIFT_;
          n32SubVStep = n32VStep << _RUNSHIFT_;
        }
        
        // Loop for number of subdivisions
        for(n32Cnt = 0; n32Cnt < n32RunLoops; n32Cnt++)
        {
          n32Len = _RUNLEN_;

          n32RZ  = n32Z - _pcsAct3DEnv_->mt_tRB->n32ZTrans;
          n32TX1 = (n32U / n32RZ) << _LSTEP_PREC_;
          n32TY1 = (n32V / n32RZ) << _LSTEP_PREC_;

          n32RZ += n32SubZStep;
          n32U  += n32SubUStep;
          n32V  += n32SubVStep;

          n32TX2 = (n32U / n32RZ) << _LSTEP_PREC_;
          n32TY2 = (n32V / n32RZ) << _LSTEP_PREC_;
          
          n32TXStep = (n32TX2 - n32TX1) >> _RUNSHIFT_;
          n32TYStep = ((n32TY2 - n32TY1) >> _RUNSHIFT_) << un32PreMultShift;
          
          n32TX = n32TX1;
          n32TY = n32TY1 << un32PreMultShift;

          // Rasterize the scanline
          while(n32Len-- > 0)
          {
            if(*pn32Z < n32Z)
            {
              *pn32Z = n32Z;

              n32Pos = ((n32TY & un32Mask) + n32TX) >> _LSTEP_PREC_ << 2;
              *pun32Buff = _pun32RFogLightTable_[(n32F >> _ISTEP_PREC_ << 10) | (n32R >> _ISTEP_PREC_ << 5) | pun8TexBuff[n32Pos + 2]] |
                           _pun32GFogLightTable_[(n32F >> _ISTEP_PREC_ << 10) | (n32G >> _ISTEP_PREC_ << 5) | pun8TexBuff[n32Pos + 1]] |
                           _pun32BFogLightTable_[(n32F >> _ISTEP_PREC_ << 10) | (n32B >> _ISTEP_PREC_ << 5) | pun8TexBuff[n32Pos + 0]];
            }
            n32Z  += n32ZStep;
            n32TX += n32TXStep;
            n32TY += n32TYStep;
            
            n32R += n32RStep;
            n32G += n32GStep;
            n32B += n32BStep;

            n32F += n32FStep;
            
            pn32Z++;
            pun32Buff++; 
          }
        }
        
        // Rasterize the remainder of the scanline
        n32Len = n32Width & (_RUNLEN_ - 1);
        if(n32Len > 0)
        {
          n32RZ  = n32Z - _pcsAct3DEnv_->mt_tRB->n32ZTrans;
          n32TX1 = (n32U / n32RZ) << _LSTEP_PREC_;
          n32TY1 = (n32V / n32RZ) << _LSTEP_PREC_;

          n32RZ += (n32ZStep * n32Len);
          n32U  += (n32UStep * n32Len);
          n32V  += (n32VStep * n32Len);

          n32TX2 = (n32U / n32RZ) << _LSTEP_PREC_;
          n32TY2 = (n32V / n32RZ) << _LSTEP_PREC_;
          
          n32TXStep = (n32TX2 - n32TX1) / n32Len;
          n32TYStep = ((n32TY2 - n32TY1) / n32Len) << un32PreMultShift;

          n32TX = n32TX1;
          n32TY = n32TY1 << un32PreMultShift;

          while(n32Len-- > 0)
          {
            if(*pn32Z < n32Z)
            {
              *pn32Z = n32Z;
                           
              n32Pos = ((n32TY & un32Mask) + n32TX) >> _LSTEP_PREC_ << 2;
              *pun32Buff = _pun32RFogLightTable_[(n32F >> _ISTEP_PREC_ << 10) | (n32R >> _ISTEP_PREC_ << 5) | pun8TexBuff[n32Pos + 2]] |
                           _pun32GFogLightTable_[(n32F >> _ISTEP_PREC_ << 10) | (n32G >> _ISTEP_PREC_ << 5) | pun8TexBuff[n32Pos + 1]] |
                           _pun32BFogLightTable_[(n32F >> _ISTEP_PREC_ << 10) | (n32B >> _ISTEP_PREC_ << 5) | pun8TexBuff[n32Pos + 0]];
            }
            n32Z  += n32ZStep;
            n32TX += n32TXStep;
            n32TY += n32TYStep;
            
            n32R += n32RStep;
            n32G += n32GStep;
            n32B += n32BStep;

            n32F += n32FStep;

            pn32Z++;
            pun32Buff++; 
          }          
        }
      }
      csRightSeg.Step();
      csLeftSeg.Step();
      un32YIndex += _pcsAct3DEnv_->mt_tRB->un32Width;
    }
  }
}

// ============================================================================================
// End of vfrndrs.cpp
// ============================================================================================
