// ############################################################################################
//  vhnode.cpp
//  Implementation of _CVortexHNode class
//  -------------------------------------------------------------------------------------------
//  Copyright (C) 2004 Aloysius Indrayanto
//  This file is part of the VortexGE
//
//  This program is free software; you can redistribute it and/or modify it under the terms of 
//  the GNU Lesser General Public License as published by the Free Software Foundation; either
//  version 2.1 of the License, or (at your option) any later version.
//
//  This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
//  without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
//  See the GNU Lesser General Public License for more details.
//
//  You should have received a copy of the GNU Lesser General Public License along with this 
//  library; if not, write to the Free Software Foundation, Inc., 59 Temple Place, Suite 330, 
//  Boston, MA 02111-1307 USA.
// ############################################################################################

// ============================================================================================
// Header files
// ============================================================================================
// ***** C++ headers
#include <string.h>
// ***** VortexGE headers
#include <vhnode.h>
#include <vmdisp.h>

// ============================================================================================
// Exported classes implementation
// ============================================================================================
// --------------------------------------------------------------------------------------------
// The hierarchical node class
// --------------------------------------------------------------------------------------------
// ***** Constructor & destructor
_CVortexHNode::_CVortexHNode()
{
  mt_pcsParent   = NULL;
  memset(mt_apcsChild, 0, sizeof(mt_apcsChild[0]) * _MAX_CHILD_);
  StcIdentity();
  DynSetIdentity();
  mt_tM.csTot.Identity();
}
_CVortexHNode::~_CVortexHNode()
{
  DelAllChild();
}

// ***** Hierarchy constructions
_vbool _CVortexHNode::SetParent(_CVortexHNode *pcsNewParent)
{
  // ..... Temporary variables
  _vuint8 un8Cnt;
  
  // ..... If the new parent is not NULL
  if(pcsNewParent)
  {
    // Search for an empty node in the new parent
    for(un8Cnt = 0; un8Cnt < _MAX_CHILD_; un8Cnt++)
      if(pcsNewParent->mt_apcsChild[un8Cnt] == NULL) break;
    if(un8Cnt >= _MAX_CHILD_)
    {
      VSetErrorCode(VERR_TooMany);
      VError(_F_, _L_, "Parent %X cannot has any more child", _vuintad(pcsNewParent));
      return(false);
    }

    // Add myself to the new parent
    pcsNewParent->mt_apcsChild[un8Cnt] = this;
  }
  
  // ..... Remove myself from the old parent (if any)
  if(mt_pcsParent)
  {
    for(un8Cnt = 0; un8Cnt < _MAX_CHILD_; un8Cnt++)
      if(mt_pcsParent->mt_apcsChild[un8Cnt] == this)
      {
        mt_pcsParent->mt_apcsChild[un8Cnt] = NULL;
        break;
      }
  }
  
  // ..... OK !
  mt_pcsParent = pcsNewParent;
  return(true);
  
}
_CVortexHNode *_CVortexHNode::AddChild(_vvoid)
{
  // ..... Temporary variables
  _CVortexHNode *pcsNew;
  _vuint8       un8Cnt;
  
  // ..... Search for an empty node
  for(un8Cnt = 0; un8Cnt < _MAX_CHILD_; un8Cnt++)
    if(mt_apcsChild[un8Cnt] == NULL) break;
  if(un8Cnt >= _MAX_CHILD_)
  {
    VSetErrorCode(VERR_TooMany);
    VError(_F_, _L_, "Each hierarchical node cannot has more than %d children", _MAX_CHILD_);
    return(NULL);
  }
  
  // ..... Allocate memory for new node
  pcsNew = new _CVortexHNode;
  if(pcsNew == NULL)
  {
    VSetErrorCode(VERR_NotEnoughMemory);
    VError(_F_, _L_, "Not enough memory to create a new hierarchical node");
    return(NULL);
  }
  
  // ..... Set properties
  pcsNew->mt_pcsParent = this;
  mt_apcsChild[un8Cnt] = pcsNew;
  
  // ..... OK
  return(pcsNew);
}
_vvoid _CVortexHNode::DelChild(_CVortexHNode *pcsChild)
{
  // ..... Temporary variables
  _vuint8 un8Cnt;
  
  // ..... Search and delete a specific child
  for(un8Cnt = 0; un8Cnt < _MAX_CHILD_; un8Cnt++)
    if(mt_apcsChild[un8Cnt] == pcsChild)
    {
      delete pcsChild;
      mt_apcsChild[un8Cnt] = NULL;
      break;
    }
}
_vvoid _CVortexHNode::DelAllChild(_vvoid)
{
  // ..... Temporary variables
  _vuint8 un8Cnt;
  
  // ..... Search and delete all childdren
  for(un8Cnt = 0; un8Cnt < _MAX_CHILD_; un8Cnt++)
    if(mt_apcsChild[un8Cnt])
    {
      delete mt_apcsChild[un8Cnt];
      mt_apcsChild[un8Cnt] = NULL;
    }
}

// ============================================================================================
// End of vhnode.cpp
// ============================================================================================
