// ############################################################################################
//  vx11dlg.h
//  Definition & implementation of dialog classes used by _CVortexVideoEnv class
//  using X Window system (Xlib)
//  -------------------------------------------------------------------------------------------
//  Copyright (C) 2004 Aloysius Indrayanto
//  This file is part of the VortexGE
//
//  This program is free software; you can redistribute it and/or modify it under the terms of 
//  the GNU Lesser General Public License as published by the Free Software Foundation; either
//  version 2.1 of the License, or (at your option) any later version.
//
//  This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
//  without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
//  See the GNU Lesser General Public License for more details.
//
//  You should have received a copy of the GNU Lesser General Public License along with this 
//  library; if not, write to the Free Software Foundation, Inc., 59 Temple Place, Suite 330, 
//  Boston, MA 02111-1307 USA.
// ############################################################################################
#ifndef _VX11DLG_H_
#define _VX11DLG_H_

// ============================================================================================
// Header files
// ============================================================================================
// ***** X11 headers 
#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <X11/keysym.h>
// ***** VortexGE headers
#include <vbtypes.h>
 
// ============================================================================================
// Exported classes definition
// ============================================================================================
// ***** The button class
class _CVortexButton
{
  protected:
    // Variables
    XSegment   mt_atSegTL[4];
    XSegment   mt_atSegBR[4];
    XRectangle mt_tFace;
    XRectangle mt_tIn;
    XRectangle mt_tOut;
    XRectangle mt_tOutThick;
    _vint32    mt_n32Xt, mt_n32Yt;
    _vuint32   mt_un32TextLen;
    _vchar     *mt_pcText;
    _vbool     mt_bInsideOutline;
    _vbool     mt_bThickOutline;
    _vbool     mt_bDuplicateString;
  
  public:
    // Constructor & destructor
    inline _CVortexButton();
    inline ~_CVortexButton();
    // Button functions
    inline _vbool MouseOver(_vint32 n32PointerX, _vint32 n32PointerY);
  
  // Friend classes
  friend class _CVortexVideoEnv;
};

// ***** The panel-bar class
class _CVortexPanelBar
{
  protected:
    // Variables
    XSegment   mt_atSegLL[6];
    XSegment   mt_atSegTL[8]; 
    XSegment   mt_atSegBR[8]; 
    XRectangle mt_atFace[2];
    XRectangle mt_atOut[2];
    _vint32    mt_n32XC, mt_n32XCW, mt_n32YC;
    _vint32    mt_n32Xt, mt_n32Yt;
    _vuint32   mt_un32CaptionLen;
    _vchar     *mt_pcCaption;
    _vbool     mt_bDuplicateString;
  
  public:
    // Constructor & destructor
    inline _CVortexPanelBar();
    inline ~_CVortexPanelBar();
  
  // Friend classes
  friend class _CVortexVideoEnv;
};

// ############################################################################################
// Inline functions implementation
// ############################################################################################
// ============================================================================================
// From exported classes
// ============================================================================================
// --------------------------------------------------------------------------------------------
// The button class
// --------------------------------------------------------------------------------------------
// ***** Constructor & destructor
_CVortexButton::_CVortexButton()
{
  mt_bDuplicateString = false;
  mt_pcText           = NULL;
}
_CVortexButton::~_CVortexButton()
{
  if((mt_bDuplicateString) && (mt_pcText != NULL)) free(mt_pcText);
}

// ***** Button functions
_vbool _CVortexButton::MouseOver(_vint32 n32PointerX, _vint32 n32PointerY)
{
  if((n32PointerX >= mt_tFace.x) && (n32PointerX < (mt_tFace.x + mt_tFace.width)) && 
     (n32PointerY >= mt_tFace.y) && (n32PointerY < (mt_tFace.y + mt_tFace.height)))
    return(true);
  return(false);
}

// --------------------------------------------------------------------------------------------
// The panel-bar class
// --------------------------------------------------------------------------------------------
// ***** Constructor & destructor
_CVortexPanelBar::_CVortexPanelBar()
{
  mt_bDuplicateString = false;
  mt_pcCaption        = NULL;
}
_CVortexPanelBar::~_CVortexPanelBar()
{
  if((mt_bDuplicateString) && (mt_pcCaption != NULL)) free(mt_pcCaption);
}

#endif
// ============================================================================================
// End of vx11dlg.h
// ============================================================================================
