// ############################################################################################
//  vmdisp.h
//  Prototype of console-message-displayer functions
//  -------------------------------------------------------------------------------------------
//  Copyright (C) 2004 Aloysius Indrayanto
//  This file is part of the VortexGE
//
//  This program is free software; you can redistribute it and/or modify it under the terms of 
//  the GNU Lesser General Public License as published by the Free Software Foundation; either
//  version 2.1 of the License, or (at your option) any later version.
//
//  This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
//  without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
//  See the GNU Lesser General Public License for more details.
//
//  You should have received a copy of the GNU Lesser General Public License along with this 
//  library; if not, write to the Free Software Foundation, Inc., 59 Temple Place, Suite 330, 
//  Boston, MA 02111-1307 USA.
// ############################################################################################
#ifndef _VMDISP_H_
#define _VMDISP_H_

// ============================================================================================
// Header files
// ============================================================================================
// ***** VortexGE headers
#include <vbtypes.h>

// ============================================================================================
// Enumerations
// ============================================================================================
// ***** Error codes
typedef enum _EVEC_ { VERR_None, VERR_NotImplemented,
                      VERR_XOpenDisplay, VERR_XCreateGC, VERR_XLoadQueryFont,
                      VERR_XCreateImage, VERR_NotFound, VERR_NotSupported, VERR_NotInitialized, 
                      VERR_NotEnoughMemory, VERR_InvalidParameter, VERR_OpenDevice, 
                      VERR_DeviceNotOpenend, VERR_GetDeviceParam, VERR_SetDeviceParam,
                      VERR_OpenFile, VERR_GetFileStatus, VERR_ReadFile, VERR_WriteFile,
                      VERR_InvalidFile, VERR_IncorrectVersion, VERR_InvalidFileHeader, 
                      VERR_InvalidHeaderParam, VERR_PackData, VERR_UnpackData, VERR_CompressData, 
                      VERR_UncompressData, VERR_CreateThread, VERR_NotResponding, 
                      VERR_MultipleInstance, VERR_TooMany
                    } _EVortexERRCode;

// ============================================================================================
// Exported functions prototype
// ============================================================================================
// ***** Message displayer
extern _vvoid VClrSrc(_vvoid);
extern _vvoid VFatalError(const _vchar *pcMsg, ...);
extern _vvoid VFatalError(const _vchar *pcFile, _vint32 n32Line, const _vchar *pcMsg, ...);
extern _vvoid VError(const _vchar *pcMsg, ...);
extern _vvoid VError(const _vchar *pcFile, _vint32 n32Line, const _vchar *pcMsg, ...);
extern _vvoid VWarning(const _vchar *pcMsg, ...);
extern _vvoid VWarning(const _vchar *pcFile, _vint32 n32Line, const _vchar *pcMsg, ...);
extern _vvoid VInfo(const _vchar *pcMsg, ...);
extern _vvoid VInfo(const _vchar *pcFile, _vint32 n32Line, const _vchar *pcMsg, ...);
extern _vvoid VPrint(const _vchar *pcMsg, ...);
extern _vvoid VRedirectMsgs(_TVortexFILE *ptInfo = stdout, _TVortexFILE *ptWarning = stdout, _TVortexFILE *ptError = stdout);
                    
// ***** Error message controller
extern _vvoid VClearError(_vvoid);
extern _vvoid VSetErrorCode(_EVortexERRCode eERRCode);
extern _EVortexERRCode GetLastErrorCode(_vvoid);
extern const _vchar *GetLastErrorString(_vvoid);

#endif
// ============================================================================================
// End of vmdisp.h
// ============================================================================================
