// ############################################################################################
//  vnhclips.cpp
//  Implementation of normal (unfogged) horizontal-line clipping functions
//  -------------------------------------------------------------------------------------------
//  Copyright (C) 2004 Aloysius Indrayanto
//  This file is part of the VortexGE
//
//  The source codes in this file is adopted from the book :
//    Cutting-Edge 3D Game Programming with C++
//    John De Goes
//    Coriolis Group, Inc.
//    USA, 1996
//
//  This program is free software; you can redistribute it and/or modify it under the terms of 
//  the GNU Lesser General Public License as published by the Free Software Foundation; either
//  version 2.1 of the License, or (at your option) any later version.
//
//  This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
//  without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
//  See the GNU Lesser General Public License for more details.
//
//  You should have received a copy of the GNU Lesser General Public License along with this 
//  library; if not, write to the Free Software Foundation, Inc., 59 Temple Place, Suite 330, 
//  Boston, MA 02111-1307 USA.
// ############################################################################################

// ============================================================================================
// Header files
// ============================================================================================
// ***** VortexGE headers
#include <vpanel3d.h>

// ============================================================================================
// Exported functions implementation
// ============================================================================================
// --------------------------------------------------------------------------------------------
// Horizontal-line clipping functions
// --------------------------------------------------------------------------------------------
// ***** Untextured, flat shaded
_vvoid _UFClipHLine_(_vint32 &n32X1, _vint32 &n32X2, _vint32 &n32Z, _vint32 &n32ZStep)
{
  if(n32X1 < _pcsAct3DEnv_->mt_n32MinX)
  {
    n32Z += (n32ZStep * (_pcsAct3DEnv_->mt_n32MinX - n32X1));
    n32X1 = _pcsAct3DEnv_->mt_n32MinX;
  }
  else if(n32X1 > _pcsAct3DEnv_->mt_n32MaxX)
    n32X1 = _pcsAct3DEnv_->mt_n32MaxX;
  
  if(n32X2 < _pcsAct3DEnv_->mt_n32MinX)
    n32X2 = _pcsAct3DEnv_->mt_n32MinX;
  else if(n32X2 > _pcsAct3DEnv_->mt_n32MaxX)
    n32X2 = _pcsAct3DEnv_->mt_n32MaxX;
}

// ***** Untextured, gouroud shaded
_vvoid _UGClipHLine_(_vint32 &n32X1, _vint32 &n32X2, _vint32 &n32Z, _vint32 &n32ZStep,
                     _vint32 &n32R, _vint32 &n32RStep, _vint32 &n32G, _vint32 &n32GStep,
                     _vint32 &n32B, _vint32 &n32BStep)
{
  if(n32X1 < _pcsAct3DEnv_->mt_n32MinX)
  {
    n32Z += (n32ZStep * (_pcsAct3DEnv_->mt_n32MinX - n32X1));
    n32X1 = _pcsAct3DEnv_->mt_n32MinX;
    n32R += (n32RStep * (_pcsAct3DEnv_->mt_n32MinX - n32X1));
    n32G += (n32GStep * (_pcsAct3DEnv_->mt_n32MinX - n32X1));
    n32B += (n32BStep * (_pcsAct3DEnv_->mt_n32MinX - n32X1));
  }
  else if(n32X1 > _pcsAct3DEnv_->mt_n32MaxX)
    n32X1 = _pcsAct3DEnv_->mt_n32MaxX;
  
  if(n32X2 < _pcsAct3DEnv_->mt_n32MinX)
    n32X2 = _pcsAct3DEnv_->mt_n32MinX;
  else if(n32X2 > _pcsAct3DEnv_->mt_n32MaxX)
    n32X2 = _pcsAct3DEnv_->mt_n32MaxX;
}

// ***** Textured, flat shaded
_vvoid _TFClipHLine_(_vint32 &n32X1, _vint32 &n32X2, _vint32 &n32Z, _vint32 &n32ZStep,
                     _vint32 &n32U, _vint32 &n32UStep, _vint32 &n32V, _vint32 &n32VStep)
{
  if(n32X1 < _pcsAct3DEnv_->mt_n32MinX)
  {
    n32Z += (n32ZStep * (_pcsAct3DEnv_->mt_n32MinX - n32X1));
    n32X1 = _pcsAct3DEnv_->mt_n32MinX;
    n32U += (n32UStep * (_pcsAct3DEnv_->mt_n32MinX - n32X1));
    n32V += (n32VStep * (_pcsAct3DEnv_->mt_n32MinX - n32X1));
  }
  else if(n32X1 > _pcsAct3DEnv_->mt_n32MaxX)
    n32X1 = _pcsAct3DEnv_->mt_n32MaxX;
  
  if(n32X2 < _pcsAct3DEnv_->mt_n32MinX)
    n32X2 = _pcsAct3DEnv_->mt_n32MinX;
  else if(n32X2 > _pcsAct3DEnv_->mt_n32MaxX)
    n32X2 = _pcsAct3DEnv_->mt_n32MaxX;
}

// ***** Textured, gouroud shaded
_vvoid _TGClipHLine_(_vint32 &n32X1, _vint32 &n32X2, _vint32 &n32Z, _vint32 &n32ZStep,
                     _vint32 &n32U, _vint32 &n32UStep, _vint32 &n32V, _vint32 &n32VStep,
                     _vint32 &n32R, _vint32 &n32RStep, _vint32 &n32G, _vint32 &n32GStep,
                     _vint32 &n32B, _vint32 &n32BStep)
{
  if(n32X1 < _pcsAct3DEnv_->mt_n32MinX)
  {
    n32Z += (n32ZStep * (_pcsAct3DEnv_->mt_n32MinX - n32X1));
    n32X1 = _pcsAct3DEnv_->mt_n32MinX;
    n32R += (n32RStep * (_pcsAct3DEnv_->mt_n32MinX - n32X1));
    n32G += (n32GStep * (_pcsAct3DEnv_->mt_n32MinX - n32X1));
    n32B += (n32BStep * (_pcsAct3DEnv_->mt_n32MinX - n32X1));
    n32U += (n32UStep * (_pcsAct3DEnv_->mt_n32MinX - n32X1));
    n32V += (n32VStep * (_pcsAct3DEnv_->mt_n32MinX - n32X1));
  }
  else if(n32X1 > _pcsAct3DEnv_->mt_n32MaxX)
    n32X1 = _pcsAct3DEnv_->mt_n32MaxX;
  
  if(n32X2 < _pcsAct3DEnv_->mt_n32MinX)
    n32X2 = _pcsAct3DEnv_->mt_n32MinX;
  else if(n32X2 > _pcsAct3DEnv_->mt_n32MaxX)
    n32X2 = _pcsAct3DEnv_->mt_n32MaxX;
}

// ============================================================================================
// End of vnhclips.cpp
// ============================================================================================
