// ############################################################################################
//  vp3dbuff.cpp
//  Implementation of _CVortexPanel3DBuffer class
//  -------------------------------------------------------------------------------------------
//  Copyright (C) 2004 Aloysius Indrayanto
//  This file is part of the VortexGE
//
//  The source codes in this file is adopted from the book :
//    Cutting-Edge 3D Game Programming with C++
//    John De Goes
//    Coriolis Group, Inc.
//    USA, 1996
//
//  This program is free software; you can redistribute it and/or modify it under the terms of 
//  the GNU Lesser General Public License as published by the Free Software Foundation; either
//  version 2.1 of the License, or (at your option) any later version.
//
//  This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
//  without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
//  See the GNU Lesser General Public License for more details.
//
//  You should have received a copy of the GNU Lesser General Public License along with this 
//  library; if not, write to the Free Software Foundation, Inc., 59 Temple Place, Suite 330, 
//  Boston, MA 02111-1307 USA.
// ############################################################################################

// ============================================================================================
// Header files
// ============================================================================================
// ***** C++ headers
#include <stdio.h>
#include <stdlib.h>
// ***** VortexGE headers
#include <vmdisp.h>
#include <vp3dbuff.h>

// ============================================================================================
// Defines
// ============================================================================================
// ***** VPD header
#define _VPD_ID1_ 'V'
#define _VPD_ID2_ 'P'
#define _VPD_ID3_ 'D'
#define _VPD_VER_ 100

// ============================================================================================
// Structures
// ============================================================================================
// ***** VPD header
struct _TVPDHeader
{ 
  _vchar   acID[3];         // File ID, must be "VPD"
  _vuint8  un8Version;      // File version
  _vuint32 un32VertexCount; // Vertex count
  _vuint32 un32PanelCount;  // Panel count
  _vuint16 un16Reserved;    // Reserved, must be zero
} __attribute__((packed));

// ============================================================================================
// Exported classes implementation
// ============================================================================================
// --------------------------------------------------------------------------------------------
// The 3d panel buffer class
// --------------------------------------------------------------------------------------------
// ***** Texturing functions
_vvoid _CVortexPanel3DBuffer::CalcDefUnitTexCoords(_vvoid)
{
  mt_un32TWidth1  = 1;
  mt_un32THeight1 = 1;
  for(_vuint32 un32Cnt = 0; un32Cnt < mt_un32PCount; un32Cnt++)
  {
    mt_pcsPList[un32Cnt].mt_tP.tUV[0].m_n32U = 0;
    mt_pcsPList[un32Cnt].mt_tP.tUV[0].m_n32V = mt_un32THeight1;
    mt_pcsPList[un32Cnt].mt_tP.tUV[1].m_n32U = 0;
    mt_pcsPList[un32Cnt].mt_tP.tUV[1].m_n32V = 0;
    mt_pcsPList[un32Cnt].mt_tP.tUV[2].m_n32U = mt_un32TWidth1;
    mt_pcsPList[un32Cnt].mt_tP.tUV[2].m_n32V = 0;
    mt_pcsPList[un32Cnt].mt_tP.tUV[3].m_n32U = mt_un32TWidth1;
    mt_pcsPList[un32Cnt].mt_tP.tUV[3].m_n32V = mt_un32THeight1;
  }
}
_vvoid _CVortexPanel3DBuffer::CalcDefTexCoords(_vvoid)
{
  mt_un32TWidth1  = mt_pcsTexture->mt_un32Width - 1;
  mt_un32THeight1 = mt_pcsTexture->mt_un32Height - 1;
  for(_vuint32 un32Cnt = 0; un32Cnt < mt_un32PCount; un32Cnt++)
  {
    mt_pcsPList[un32Cnt].mt_tP.tUV[0].m_n32U = 0;
    mt_pcsPList[un32Cnt].mt_tP.tUV[0].m_n32V = mt_un32THeight1;
    mt_pcsPList[un32Cnt].mt_tP.tUV[1].m_n32U = 0;
    mt_pcsPList[un32Cnt].mt_tP.tUV[1].m_n32V = 0;
    mt_pcsPList[un32Cnt].mt_tP.tUV[2].m_n32U = mt_un32TWidth1;
    mt_pcsPList[un32Cnt].mt_tP.tUV[2].m_n32V = 0;
    mt_pcsPList[un32Cnt].mt_tP.tUV[3].m_n32U = mt_un32TWidth1;
    mt_pcsPList[un32Cnt].mt_tP.tUV[3].m_n32V = mt_un32THeight1;
  }
}
_vvoid _CVortexPanel3DBuffer::RefitTexCoords(_vvoid)
{
  // ..... Temporary variables
  _vfloat  fScaleU, fScaleV;
  _vuint32 un32Cnt1, un32Cnt2;
  _vint32  n32U, n32V;
    
  // ..... Check
  if((mt_un32TWidth1 == (mt_pcsTexture->mt_un32Width - 1)) && 
     (mt_un32THeight1 == (mt_pcsTexture->mt_un32Height - 1))) return;
  
  // ..... Calculate scalling
  fScaleU = _vfloat(mt_pcsTexture->mt_un32Width - 1) / _vfloat(mt_un32TWidth1);
  fScaleV = _vfloat(mt_pcsTexture->mt_un32Height - 1) / _vfloat(mt_un32THeight1);

  // ..... Refit texture
  mt_un32TWidth1  = mt_pcsTexture->mt_un32Width - 1;
  mt_un32THeight1 = mt_pcsTexture->mt_un32Height - 1;
  for(un32Cnt1 = 0; un32Cnt1 < mt_un32PCount; un32Cnt1++)
  {
    for(un32Cnt2 = 0; un32Cnt2 < 4; un32Cnt2++)
    {
      n32U = mt_pcsPList[un32Cnt1].mt_tP.tUV[un32Cnt2].m_n32U;
      n32V = mt_pcsPList[un32Cnt1].mt_tP.tUV[un32Cnt2].m_n32V;
      n32U = _vint32(_vfloat(n32U) * fScaleU);
      n32V = _vint32(_vfloat(n32V) * fScaleV);
      if(n32U > _vint32(mt_un32TWidth1)) n32U = mt_un32TWidth1;
      if(n32V > _vint32(mt_un32THeight1)) n32V = mt_un32THeight1;
      mt_pcsPList[un32Cnt1].mt_tP.tUV[un32Cnt2].m_n32U = n32U;
      mt_pcsPList[un32Cnt1].mt_tP.tUV[un32Cnt2].m_n32V = n32V;
    }
  }
}

// ***** Builder functions
_vbool _CVortexPanel3DBuffer::InitFromMemory(_vfloat *pfVertices, _vuint32 un32VertexCount, 
                                             _vuint32 *pfPanelPoints, _vuint32 un32PanelCount,
                                             _TVortexAmbientLight *ptVAmbientLight, 
                                             _TVortexOmniLight *ptVOmniLight,
                                             _vbool bEnableSmoothShading)
{
  // ..... Temporary variables
  _vuint32 un32Cnt, un32Index;
  
  // ..... Destroy previous
  if(mt_pcsVList) delete [] mt_pcsVList;
  if(mt_pcsPList) delete [] mt_pcsPList;
  mt_pcsVList   = NULL;
  mt_pcsPList   = NULL;
  mt_un32VCount = 0;
  mt_un32PCount = 0;

  // ..... Allocate memory
  mt_pcsVList = new _CVortexPoint3D[un32VertexCount];
  if(mt_pcsVList == NULL)
  {
    VSetErrorCode(VERR_NotEnoughMemory);
    VError(_F_, _L_, "Failed to allocate memory for vertex buffer");
    return(false);
  }
  mt_pcsPList = new _CVortexPanel3D[un32PanelCount];
  if(mt_pcsPList == NULL)
  {
    delete [] mt_pcsVList;
    mt_pcsVList = NULL;

    VSetErrorCode(VERR_NotEnoughMemory);
    VError(_F_, _L_, "Failed to allocate memory for panel buffer");
    return(false);
  }
  
  // ..... Copy vertex data
  un32Index = 0;
  for(un32Cnt = 0; un32Cnt < un32VertexCount; un32Cnt++)
  {
    mt_pcsVList[un32Cnt].m_tLocal.fX = pfVertices[un32Index++];
    mt_pcsVList[un32Cnt].m_tLocal.fY = pfVertices[un32Index++];
    mt_pcsVList[un32Cnt].m_tLocal.fZ = pfVertices[un32Index++];
    //if(fabs(mt_pcsVList[un32Cnt].m_tLocal.fZ) < _vfloat(0.00001)) mt_pcsVList[un32Cnt].m_tLocal.fZ = _vfloat(0.00001);
  }
  
  // ..... Generate panel data
  un32Index = 0;
  for(un32Cnt = 0; un32Cnt < un32PanelCount; un32Cnt++)
  {
    mt_pcsPList[un32Cnt].mt_tP.csVPoint[0] = &mt_pcsVList[pfPanelPoints[un32Index++]];
    mt_pcsPList[un32Cnt].mt_tP.csVPoint[1] = &mt_pcsVList[pfPanelPoints[un32Index++]];
    mt_pcsPList[un32Cnt].mt_tP.csVPoint[2] = &mt_pcsVList[pfPanelPoints[un32Index++]];
    mt_pcsPList[un32Cnt].mt_tP.csVPoint[3] = &mt_pcsVList[pfPanelPoints[un32Index++]];
    
    mt_pcsPList[un32Cnt]._CalcNormal_();
    mt_pcsPList[un32Cnt]._CalcIntenAmbient_(ptVAmbientLight);
    mt_pcsPList[un32Cnt]._FSCalcIntenOmni_(ptVOmniLight);
    mt_pcsPList[un32Cnt]._FSCalcInten_();
  }
  if(bEnableSmoothShading)
  {
    for(un32Cnt = 0; un32Cnt < un32PanelCount; un32Cnt++)
    {
      mt_pcsPList[un32Cnt]._GSCalcIntenOmni_(ptVOmniLight);
      mt_pcsPList[un32Cnt]._GSCalcInten_();
    }
  }
  
  // ..... OK !
  mt_un32VCount = un32VertexCount;
  mt_un32PCount = un32PanelCount;
  return(true);
}

// ***** File I/O functions
_vbool _CVortexPanel3DBuffer::SaveVPD(const _vchar *pcFileName)
{
  // ..... Temporary variables
  _TVortexFILE *ptFile;
  _TVPDHeader  tHeader;
  _vuint32     un32Cnt, un32Cnt2;
  _vchar       acFile[256];
  
  // ..... Check
  if((mt_un32VCount == 0) || (mt_un32PCount == 0))
  { 
    VSetErrorCode(VERR_NotInitialized);
    VError(_F_, _L_, "No data to save (this _CVortexPanel3DBuffer is not initialized)");
    return(false);
  }
  
  // ..... Open file
  strcpy(acFile, pcFileName);
  if(strstr(acFile, ".vpf") == NULL) strcat(acFile, ".vpf");
  ptFile = fopen(acFile, "wb");
  if(ptFile == NULL)
  {
    VSetErrorCode(VERR_OpenFile);
    VError(_F_, _L_, "Failed opening file \"%s\"", acFile);
    return(false);
  }    

  // ..... Make header
  memset(&tHeader, 0, sizeof(tHeader));
  tHeader.acID[0]         = _VPD_ID1_;
  tHeader.acID[1]         = _VPD_ID2_;
  tHeader.acID[2]         = _VPD_ID3_;
  tHeader.un8Version      = _VPD_VER_;
  tHeader.un32VertexCount = mt_un32VCount;
  tHeader.un32PanelCount  = mt_un32PCount;

  // ..... Write header
  if(fwrite(&tHeader, sizeof(tHeader), 1, ptFile) < 1) goto ERRWrite;

  // ..... Write vertex data
  for(un32Cnt = 0; un32Cnt < mt_un32VCount; un32Cnt++)
  {
    if(fwrite(&mt_pcsVList[un32Cnt].m_tLocal.fX, sizeof(mt_pcsVList[un32Cnt].m_tLocal.fX), 1, ptFile) < 1) goto ERRWrite;
    if(fwrite(&mt_pcsVList[un32Cnt].m_tLocal.fY, sizeof(mt_pcsVList[un32Cnt].m_tLocal.fY), 1, ptFile) < 1) goto ERRWrite;
    if(fwrite(&mt_pcsVList[un32Cnt].m_tLocal.fZ, sizeof(mt_pcsVList[un32Cnt].m_tLocal.fZ), 1, ptFile) < 1) goto ERRWrite;
  }  
  
  // ..... Write panel data
  for(un32Cnt = 0; un32Cnt < mt_un32PCount; un32Cnt++)
  {
    // 1st vertex
    for(un32Cnt2 = 0; un32Cnt2 < mt_un32VCount; un32Cnt2++)
      if(mt_pcsPList[un32Cnt].mt_tP.csVPoint[0] == &mt_pcsVList[un32Cnt2])
      {
        if(fwrite(&un32Cnt2, sizeof(un32Cnt2), 1, ptFile) < 1) goto ERRWrite;
        break;
      }

    // 2nd vertex
    for(un32Cnt2 = 0; un32Cnt2 < mt_un32VCount; un32Cnt2++)
      if(mt_pcsPList[un32Cnt].mt_tP.csVPoint[1] == &mt_pcsVList[un32Cnt2])
      {
        if(fwrite(&un32Cnt2, sizeof(un32Cnt2), 1, ptFile) < 1) goto ERRWrite;
        break;
      }
      
    // 3rd vertex
    for(un32Cnt2 = 0; un32Cnt2 < mt_un32VCount; un32Cnt2++)
      if(mt_pcsPList[un32Cnt].mt_tP.csVPoint[2] == &mt_pcsVList[un32Cnt2])
      {
        if(fwrite(&un32Cnt2, sizeof(un32Cnt2), 1, ptFile) < 1) goto ERRWrite;
        break;
      }
      
    // 4th vertex
    for(un32Cnt2 = 0; un32Cnt2 < mt_un32VCount; un32Cnt2++)
      if(mt_pcsPList[un32Cnt].mt_tP.csVPoint[3] == &mt_pcsVList[un32Cnt2])
      {
        if(fwrite(&un32Cnt2, sizeof(un32Cnt2), 1, ptFile) < 1) goto ERRWrite;
        break;
      }
  }

  // ..... OK !
  fclose(ptFile);
  return(true);
  
  // ..... Error writing file
ERRWrite:;
  fclose(ptFile);
  VSetErrorCode(VERR_WriteFile);
  VError(_F_, _L_, "Failed writing file \"%s\"", acFile);
  return(false);
}
_vbool _CVortexPanel3DBuffer::LoadVPD(const _vchar *pcFileName)
{
  // ..... Temporary variables
  _TVortexFILE *ptFile;
  _TVPDHeader  tHeader;
  _vuint32     *pun32Panels;
  _vfloat      *pfVertices;
  _vchar       acFile[256];
  _vbool       bR;
 
  // ..... Open file
  strcpy(acFile, pcFileName);
  if(strstr(acFile, ".vpf") == NULL) strcat(acFile, ".vpf");
  ptFile = fopen(acFile, "rb");
  if(ptFile == NULL)
  {
    VSetErrorCode(VERR_OpenFile);
    VError(_F_, _L_, "Failed opening file \"%s\"", acFile);
    return(false);
  }    

  // ..... Read header
  if(fread(&tHeader, sizeof(tHeader), 1, ptFile) < 1) goto ERRRead;  

  // ..... Check header
  if((tHeader.acID[0] != _VPD_ID1_) || (tHeader.acID[1] != _VPD_ID2_) || (tHeader.acID[2] != _VPD_ID3_))
  {
    fclose(ptFile);
    VSetErrorCode(VERR_InvalidFile);
    VError(_F_, _L_, "\"%s\" is an invalid VPD file", acFile);
    return(false);
  }
  if(tHeader.un8Version != _VPD_VER_)
  {
    fclose(ptFile);
    VSetErrorCode(VERR_IncorrectVersion);
    VError(_F_, _L_, "\"%s\" has an incorrect version", acFile);
    return(false);
  }
  if(tHeader.un16Reserved)
  {
    fclose(ptFile);
    VSetErrorCode(VERR_InvalidFileHeader);
    VError(_F_, _L_, "Invalid header data found in file \"%s\"", acFile);
    return(false);
  }

  // ..... Allocate temporary memory
  pfVertices = (_vfloat *) malloc(tHeader.un32VertexCount* 3 * sizeof(_vfloat));
  if(pfVertices == NULL)
  {
    fclose(ptFile);
    VSetErrorCode(VERR_NotEnoughMemory);
    VError(_F_, _L_, "Failed to allocate memory for loading vertex data");
    return(false);
  }    
  pun32Panels = (_vuint32 *) malloc(tHeader.un32PanelCount * 4 * sizeof(_vuint32));
  if(pun32Panels == NULL)
  {
    free(pfVertices);
    fclose(ptFile);
    VSetErrorCode(VERR_NotEnoughMemory);
    VError(_F_, _L_, "Failed to allocate memory for loading panel data");
    return(false);
  }
  
  // ..... Read vertex and panel data
  if(fread(pfVertices, sizeof(_vfloat), tHeader.un32VertexCount * 3, ptFile) < (tHeader.un32VertexCount * 3)) goto ERRRead;
  if(fread(pun32Panels, sizeof(_vuint32), tHeader.un32PanelCount * 4, ptFile) < (tHeader.un32PanelCount * 4)) goto ERRRead;
    
  // ..... Close file
  fclose(ptFile);
  
  // .....  Generate panels
  bR = InitFromMemory(pfVertices, tHeader.un32VertexCount, pun32Panels, tHeader.un32PanelCount);
  free(pfVertices);
  free(pun32Panels);

  // ..... OK !
  return(bR);
  
  // ..... Error reading file
ERRRead:;
  fclose(ptFile);
  VSetErrorCode(VERR_ReadFile);
  VError(_F_, _L_, "Failed reading file \"%s\"", acFile);
  return(false);
}

// ============================================================================================
// End of v3dbuff.cpp
// ============================================================================================
