// ############################################################################################
//  vpanel3d.cpp
//  Implementation of _CVortexPanel3D class
//  -------------------------------------------------------------------------------------------
//  Copyright (C) 2004 Aloysius Indrayanto
//  This file is part of the VortexGE
//
//  The source codes in this file is adopted from the book :
//    Cutting-Edge 3D Game Programming with C++
//    John De Goes
//    Coriolis Group, Inc.
//    USA, 1996
//
//  This program is free software; you can redistribute it and/or modify it under the terms of 
//  the GNU Lesser General Public License as published by the Free Software Foundation; either
//  version 2.1 of the License, or (at your option) any later version.
//
//  This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
//  without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
//  See the GNU Lesser General Public License for more details.
//
//  You should have received a copy of the GNU Lesser General Public License along with this 
//  library; if not, write to the Free Software Foundation, Inc., 59 Temple Place, Suite 330, 
//  Boston, MA 02111-1307 USA.
// ############################################################################################

// ============================================================================================
// Header files
// ============================================================================================
// ***** VortexGE headers
#include <vluts.h>
#include <vfcline.h>
#include <vncline.h>
#include <vpanel3d.h>

// ============================================================================================
// External functions prototpes
// ============================================================================================
// ***** Normal clipping functions
_vvoid _UFClipHLine_(_vint32 &n32X1, _vint32 &n32X2, _vint32 &n32Z, _vint32 &n32ZStep);
_vvoid _UGClipHLine_(_vint32 &n32X1, _vint32 &n32X2, _vint32 &n32Z, _vint32 &n32ZStep,
                     _vint32 &n32R, _vint32 &n32RStep, _vint32 &n32G, _vint32 &n32GStep,
                     _vint32 &n32B, _vint32 &n32BStep);
_vvoid _TFClipHLine_(_vint32 &n32X1, _vint32 &n32X2, _vint32 &n32Z, _vint32 &n32ZStep,
                     _vint32 &n32U, _vint32 &n32UStep, _vint32 &n32V, _vint32 &n32VStep);
_vvoid _TGClipHLine_(_vint32 &n32X1, _vint32 &n32X2, _vint32 &n32Z, _vint32 &n32ZStep,
                     _vint32 &n32U, _vint32 &n32UStep, _vint32 &n32V, _vint32 &n32VStep,
                     _vint32 &n32R, _vint32 &n32RStep, _vint32 &n32G, _vint32 &n32GStep,
                     _vint32 &n32B, _vint32 &n32BStep);

// ***** Fogged clipping functions
_vvoid _FogUFClipHLine_(_vint32 &n32X1, _vint32 &n32X2, _vint32 &n32Z, _vint32 &n32ZStep,
                        _vint32 &n32F, _vint32 &n32FStep);
_vvoid _FogUGClipHLine_(_vint32 &n32X1, _vint32 &n32X2, _vint32 &n32Z, _vint32 &n32ZStep,
                        _vint32 &n32R, _vint32 &n32RStep, _vint32 &n32G, _vint32 &n32GStep,
                        _vint32 &n32B, _vint32 &n32BStep, _vint32 &n32F, _vint32 &n32FStep);
_vvoid _FogTFClipHLine_(_vint32 &n32X1, _vint32 &n32X2, _vint32 &n32Z, _vint32 &n32ZStep,
                        _vint32 &n32U, _vint32 &n32UStep, _vint32 &n32V, _vint32 &n32VStep,
                        _vint32 &n32F, _vint32 &n32FStep);
_vvoid _FogTGClipHLine_(_vint32 &n32X1, _vint32 &n32X2, _vint32 &n32Z, _vint32 &n32ZStep,
                        _vint32 &n32U, _vint32 &n32UStep, _vint32 &n32V, _vint32 &n32VStep,
                        _vint32 &n32R, _vint32 &n32RStep, _vint32 &n32G, _vint32 &n32GStep,
                        _vint32 &n32B, _vint32 &n32BStep, _vint32 &n32F, _vint32 &n32FStep);

// ============================================================================================
// Exported classes implementation
// ============================================================================================
// --------------------------------------------------------------------------------------------
// The 3d panel class
// --------------------------------------------------------------------------------------------
// ***** Panel functions
_vvoid _CVortexPanel3D::_CalcNormal_(_vvoid)
{
  // ..... Temporary variables
  _CVortexPoint3D acsUnique[4], csTmp;
  _vfloat         fX1, fY1, fZ1, fX2, fY2, fZ2, fX3, fY3, fZ3, fA, fB, fC, fDistance;
  _vuint8         un8Cnt, un8Range = 0;

  // ..... Create a list of unique vertices
  for(un8Cnt = 0; un8Cnt < 4; un8Cnt++)
  {
    csTmp = *mt_tP.csVPoint[un8Cnt];
    if(un8Range == 0)
      acsUnique[un8Range++] = csTmp;
    else if(UniqueVert(csTmp, acsUnique, un8Range))
      acsUnique[un8Range++] = csTmp;
  }

  // ..... Ccopy values for simplicity in the next equations
  fX1 = acsUnique[0].m_tLocal.fX;
  fY1 = acsUnique[0].m_tLocal.fY;
  fZ1 = acsUnique[0].m_tLocal.fZ;
  fX2 = acsUnique[1].m_tLocal.fX;
  fY2 = acsUnique[1].m_tLocal.fY;
  fZ2 = acsUnique[1].m_tLocal.fZ;
  fX3 = acsUnique[2].m_tLocal.fX;
  fY3 = acsUnique[2].m_tLocal.fY;
  fZ3 = acsUnique[2].m_tLocal.fZ;
  
  // ..... Use cross product to determine plane's orientation
  fA = fY1 * (fZ2 - fZ3) + fY2 * (fZ3 - fZ1) + fY3 * (fZ1 - fZ2);
  fB = fZ1 * (fX2 - fX3) + fZ2 * (fX3 - fX1) + fZ3 * (fX1 - fX2);
  fC = fX1 * (fY2 - fY3) + fX2 * (fY3 - fY1) + fX3 * (fY1 - fY2);

  // .....Get the distance to the vector
  fDistance = sqrt(fA * fA + fB * fB + fC * fC);
  
  // ..... Create the normalized normal
  mt_tP.tNormal.fX = fA / fDistance;
  mt_tP.tNormal.fY = fB / fDistance;
  mt_tP.tNormal.fZ = fC / fDistance;
}
_vvoid _CVortexPanel3D::_CalcRadius_(_vvoid)
{
  // ..... Temporary variables
  _CVortexPoint3D acsTmp[4], csCenter;
  _vfloat         afDist[4];
  _vuint8         un8Cnt;
  
  // ..... Create a temporary vertex list & calculate center of panel
  for(un8Cnt = 0; un8Cnt < 4; un8Cnt++)
  {
    acsTmp[un8Cnt] = *mt_tP.csVPoint[un8Cnt];
    csCenter      += acsTmp[un8Cnt];
  }
  csCenter /= _vfloat(4.0);
  
  // ..... Calculate the distance to each of the vertices
  for(un8Cnt = 0; un8Cnt < 4; un8Cnt++)
  {
    acsTmp[un8Cnt] -= csCenter;
    afDist[un8Cnt]  = acsTmp[un8Cnt].Mag();
  }
  
  // ..... Determine the maximum distance as the radius
  mt_tP.fRadius = afDist[0];
  for(un8Cnt = 0; un8Cnt < 4; un8Cnt++)
    if(afDist[un8Cnt] > mt_tP.fRadius) mt_tP.fRadius = afDist[un8Cnt];
}

// ***** Lighting functions
_vvoid _CVortexPanel3D::_FSCalcIntenOmni_(_TVortexOmniLight *ptVOmniLight)
{
  // ..... Temporary variables
  _vfloat fCos;
  _vfloat afXYZ[4];

  // ..... Calculate the center of panel
  afXYZ[0] = mt_tP.csVPoint[0]->m_tLocal.fX;
  afXYZ[1] = mt_tP.csVPoint[0]->m_tLocal.fY;
  afXYZ[2] = mt_tP.csVPoint[0]->m_tLocal.fZ;
  afXYZ[0] += mt_tP.csVPoint[1]->m_tLocal.fX;
  afXYZ[1] += mt_tP.csVPoint[1]->m_tLocal.fY;
  afXYZ[2] += mt_tP.csVPoint[1]->m_tLocal.fZ;
  afXYZ[0] += mt_tP.csVPoint[2]->m_tLocal.fX;
  afXYZ[1] += mt_tP.csVPoint[2]->m_tLocal.fY;
  afXYZ[2] += mt_tP.csVPoint[2]->m_tLocal.fZ;
  afXYZ[0] += mt_tP.csVPoint[3]->m_tLocal.fX;
  afXYZ[1] += mt_tP.csVPoint[3]->m_tLocal.fY;
  afXYZ[2] += mt_tP.csVPoint[3]->m_tLocal.fZ;
  afXYZ[0] /= 4; afXYZ[1] /= 4; afXYZ[2] /= 4;

  // ..... Use the supplied omni light
  if(ptVOmniLight)
  {
    fCos = VCalcLightCosValue(&mt_tP.tNormal.fX, afXYZ, &ptVOmniLight->fX);    
    mt_tP.tOmni.un8R = _vuint8(ptVOmniLight->un8R * fCos);
    mt_tP.tOmni.un8G = _vuint8(ptVOmniLight->un8G * fCos);
    mt_tP.tOmni.un8B = _vuint8(ptVOmniLight->un8B * fCos);
  }
  // ..... Use the default omni light from the active 3D environment
  else
  {
    fCos = VCalcLightCosValue(&mt_tP.tNormal.fX, afXYZ, &_pcsAct3DEnv_->mt_tDefOLight.fX);    
    mt_tP.tOmni.un8R = _vuint8(_pcsAct3DEnv_->mt_tDefOLight.un8R * fCos);
    mt_tP.tOmni.un8G = _vuint8(_pcsAct3DEnv_->mt_tDefOLight.un8G * fCos);
    mt_tP.tOmni.un8B = _vuint8(_pcsAct3DEnv_->mt_tDefOLight.un8B * fCos);
  }
}
_vvoid _CVortexPanel3D::_GSCalcIntenOmni_(_TVortexOmniLight *ptVOmniLight)
{
  // ..... Temporary variables
  _vfloat fCos;
  _vuint8 un8Cnt;

  // ..... Use the supplied omni light
  if(ptVOmniLight)
  {
    for(un8Cnt = 0; un8Cnt < 4; un8Cnt++)
    {
      fCos = VCalcLightCosValue(&mt_tP.tNormal.fX, &mt_tP.csVPoint[un8Cnt]->m_tLocal.fX, &ptVOmniLight->fX);    
      mt_tP.atOmni[un8Cnt].un8R = _vuint8(ptVOmniLight->un8R * fCos);
      mt_tP.atOmni[un8Cnt].un8G = _vuint8(ptVOmniLight->un8G * fCos);
      mt_tP.atOmni[un8Cnt].un8B = _vuint8(ptVOmniLight->un8B * fCos);
    }
  }
  // ..... Use the default omni light from the active 3D environment
  else
  {
    for(un8Cnt = 0; un8Cnt < 4; un8Cnt++)
    {
      fCos = VCalcLightCosValue(&mt_tP.tNormal.fX, &mt_tP.csVPoint[un8Cnt]->m_tLocal.fX, &_pcsAct3DEnv_->mt_tDefOLight.fX);    
      mt_tP.atOmni[un8Cnt].un8R = _vuint8(_pcsAct3DEnv_->mt_tDefOLight.un8R * fCos);
      mt_tP.atOmni[un8Cnt].un8G = _vuint8(_pcsAct3DEnv_->mt_tDefOLight.un8G * fCos);
      mt_tP.atOmni[un8Cnt].un8B = _vuint8(_pcsAct3DEnv_->mt_tDefOLight.un8B * fCos);
    }
  }
}
_vvoid _CVortexPanel3D::_DynFSCalcIntenOmni_(_TVortexOmniLight *ptVOmniLight)
{
  // ..... Temporary variables
  _vfloat fCos;
  _vfloat afXYZ[4];

  // ..... Calculate the center of panel
  afXYZ[0] = mt_tP.csVPoint[0]->m_tWorld.fX;
  afXYZ[1] = mt_tP.csVPoint[0]->m_tWorld.fY;
  afXYZ[2] = mt_tP.csVPoint[0]->m_tWorld.fZ;
  afXYZ[0] += mt_tP.csVPoint[1]->m_tWorld.fX;
  afXYZ[1] += mt_tP.csVPoint[1]->m_tWorld.fY;
  afXYZ[2] += mt_tP.csVPoint[1]->m_tWorld.fZ;
  afXYZ[0] += mt_tP.csVPoint[2]->m_tWorld.fX;
  afXYZ[1] += mt_tP.csVPoint[2]->m_tWorld.fY;
  afXYZ[2] += mt_tP.csVPoint[2]->m_tWorld.fZ;
  afXYZ[0] += mt_tP.csVPoint[3]->m_tWorld.fX;
  afXYZ[1] += mt_tP.csVPoint[3]->m_tWorld.fY;
  afXYZ[2] += mt_tP.csVPoint[3]->m_tWorld.fZ;
  afXYZ[0] /= 4; afXYZ[1] /= 4; afXYZ[2] /= 4;

  // ..... Use the supplied omni light
  if(ptVOmniLight)
  {
    fCos = VCalcLightCosValue(&mt_tP.tNormal.fX, afXYZ, &ptVOmniLight->fX);    
    mt_tP.tOmni.un8R = _vuint8(ptVOmniLight->un8R * fCos);
    mt_tP.tOmni.un8G = _vuint8(ptVOmniLight->un8G * fCos);
    mt_tP.tOmni.un8B = _vuint8(ptVOmniLight->un8B * fCos);
  }
  // ..... Use the default omni light from the active 3D environment
  else
  {
    fCos = VCalcLightCosValue(&mt_tP.tNormal.fX, afXYZ, &_pcsAct3DEnv_->mt_tDefOLight.fX);    
    mt_tP.tOmni.un8R = _vuint8(_pcsAct3DEnv_->mt_tDefOLight.un8R * fCos);
    mt_tP.tOmni.un8G = _vuint8(_pcsAct3DEnv_->mt_tDefOLight.un8G * fCos);
    mt_tP.tOmni.un8B = _vuint8(_pcsAct3DEnv_->mt_tDefOLight.un8B * fCos);
  }
}
_vvoid _CVortexPanel3D::_DynGSCalcIntenOmni_(_TVortexOmniLight *ptVOmniLight)
{
  // ..... Temporary variables
  _vfloat fCos;
  _vuint8 un8Cnt;

  // ..... Use the supplied omni light
  if(ptVOmniLight)
  {
    for(un8Cnt = 0; un8Cnt < 4; un8Cnt++)
    {
      fCos = VCalcLightCosValue(&mt_tP.tNormal.fX, &mt_tP.csVPoint[un8Cnt]->m_tWorld.fX, &ptVOmniLight->fX);    
      mt_tP.atOmni[un8Cnt].un8R = _vuint8(ptVOmniLight->un8R * fCos);
      mt_tP.atOmni[un8Cnt].un8G = _vuint8(ptVOmniLight->un8G * fCos);
      mt_tP.atOmni[un8Cnt].un8B = _vuint8(ptVOmniLight->un8B * fCos);
    }
  }
  // ..... Use the default omni light from the active 3D environment
  else
  {
    for(un8Cnt = 0; un8Cnt < 4; un8Cnt++)
    {
      fCos = VCalcLightCosValue(&mt_tP.tNormal.fX, &mt_tP.csVPoint[un8Cnt]->m_tWorld.fX, &_pcsAct3DEnv_->mt_tDefOLight.fX);    
      mt_tP.atOmni[un8Cnt].un8R = _vuint8(_pcsAct3DEnv_->mt_tDefOLight.un8R * fCos);
      mt_tP.atOmni[un8Cnt].un8G = _vuint8(_pcsAct3DEnv_->mt_tDefOLight.un8G * fCos);
      mt_tP.atOmni[un8Cnt].un8B = _vuint8(_pcsAct3DEnv_->mt_tDefOLight.un8B * fCos);
    }
  }
}

// ============================================================================================
// End of vpanel3d.cpp
// ============================================================================================
