// ############################################################################################
//  vpanel3d.h
//  Definition & implementation of _CVortexPanel3D class
//  -------------------------------------------------------------------------------------------
//  Copyright (C) 2004 Aloysius Indrayanto
//  This file is part of the VortexGE
//
//  The source codes in this file is adopted from the book :
//    Cutting-Edge 3D Game Programming with C++
//    John De Goes
//    Coriolis Group, Inc.
//    USA, 1996
//
//  This program is free software; you can redistribute it and/or modify it under the terms of 
//  the GNU Lesser General Public License as published by the Free Software Foundation; either
//  version 2.1 of the License, or (at your option) any later version.
//
//  This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
//  without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
//  See the GNU Lesser General Public License for more details.
//
//  You should have received a copy of the GNU Lesser General Public License along with this 
//  library; if not, write to the Free Software Foundation, Inc., 59 Temple Place, Suite 330, 
//  Boston, MA 02111-1307 USA.
// ############################################################################################
#ifndef _VPANEL3D_H_
#define _VPANEL3D_H_

// ============================================================================================
// Header files
// ============================================================================================
// ***** VortexGE headers
#include <v3denv.h>
#include <vmat3d.h>
#include <vtexuv.h>
#include <vvect3d.h>
#include <vpoint2d.h>
#include <vpoint3d.h>

// ============================================================================================
// Exported classes definition
// ============================================================================================
// ***** The 3D panel class
class _CVortexPanel3D
{
  protected:
    // Variables
    struct _TPanelData
    {
      _CVortexPoint3D       *csVPoint[4];  // Pointers to vertex list
      _TVortexTexUV         tUV[4];        // The texture coordinates
      _TVortexPoint2D       tSPoint[5];    // Maximum 5 on-screen points
      _TVortexVector3D _sse tNormal;       // The panel normal
      _vfloat               fRadius;       // The panel's radius
      _vuint8               un8SPCount;    // Screen point count
      _TVortexUN8RGB        tAmbient;      // The panel's color (from ambient light)
      _TVortexUN8RGB        tOmni;         // The panel's color (from omni light)
      _TVortexUN8RGB        tTotal;        // The panel's color (total)
      _TVortexUN8RGB        atOmni[4];     // The vertex's color (from omni light)
      _TVortexUN8RGB        atTotal[4];    // The vertex's color (total)
      _vuint32              un32Invisible; // Invisible flag
    } mt_tP;
    // Panel functions
    _vvoid _CalcNormal_(_vvoid);
    _vvoid _CalcRadius_(_vvoid);
    // Lighting functions
    _vvoid _FSCalcIntenOmni_(_TVortexOmniLight *ptVOmniLight);
    _vvoid _GSCalcIntenOmni_(_TVortexOmniLight *ptVOmniLight);
    _vvoid _DynFSCalcIntenOmni_(_TVortexOmniLight *ptVOmniLight);
    _vvoid _DynGSCalcIntenOmni_(_TVortexOmniLight *ptVOmniLight);
    inline _vvoid _CalcIntenAmbient_(_TVortexAmbientLight *ptVAmbientLight);
    inline _vvoid _FSCalcInten_(_vvoid);
    inline _vvoid _GSCalcInten_(_vvoid);
    
    // Rendering functions
    inline _vbool _CalcBFace_(_vvoid);
    inline _vbool _CheckExtents_(_vvoid);
    inline _vbool _HasVertex_(_CVortexPoint3D *pcsVPoint3D);
    
  public:
    // Constructor
    inline _CVortexPanel3D();
    // Clipping functions
    inline _vvoid UpdateFrame(_vvoid);
    inline _vuint32 Invisible(_vvoid);
    inline _vbool IsVisible3D(_vvoid);
    inline _vbool IsVisible2D(_vvoid);
    // Normal projection functions
    _vvoid UFProject(_vvoid); // Untextured/uncolored, flat shaded
    _vvoid UGProject(_vvoid); // Untextured/uncolored, gouroud shaded
    _vvoid TFProject(_vvoid); // Textured,           not/flat shaded
    _vvoid TGProject(_vvoid); // Textured,           gouroud shaded
    // Fogged projection functions
    _vvoid FogUFProject(_vvoid); // Fogged, untextured/uncolored, flat shaded
    _vvoid FogUGProject(_vvoid); // Fogged, untextured/uncolored, gouroud shaded
    _vvoid FogTFProject(_vvoid); // Fogged, textured,           not/flat shaded
    _vvoid FogTGProject(_vvoid); // Fogged, textured,           gouroud shaded
    // Normal rendering functions
    _vvoid UFRender(_vvoid);                                            // Untextured, flat shaded
    _vvoid UGRender(_vvoid);                                            // Untextured, gouroud shaded
    _vvoid CFRender(_vuint32 &un32R, _vuint32 &un32G, _vuint32 &un32B); // Colored,    flat shaded
    _vvoid CGRender(_vuint32 &un32R, _vuint32 &un32G, _vuint32 &un32B); // Colored,    gouroud shaded
    _vvoid TNRender(_CVortexTexture *pcsTexture);                       // Textured,   not shaded
    _vvoid TFRender(_CVortexTexture *pcsTexture);                       // Textured,   flat shaded
    _vvoid TGRender(_CVortexTexture *pcsTexture);                       // Textured,   gouroud shaded
    // Fogged rendering functions
    _vvoid FogUFRender(_vvoid);                                            // Fogged, untextured, flat shaded
    _vvoid FogUGRender(_vvoid);                                            // Fogged, untextured, gouroud shaded
    _vvoid FogCFRender(_vuint32 &un32R, _vuint32 &un32G, _vuint32 &un32B); // Fogged, colored,    flat shaded
    _vvoid FogCGRender(_vuint32 &un32R, _vuint32 &un32G, _vuint32 &un32B); // Fogged, colored,    gouroud shaded
    _vvoid FogTNRender(_CVortexTexture *pcsTexture);                       // Fogged, textured,   not shaded
    _vvoid FogTFRender(_CVortexTexture *pcsTexture);                       // Fogged, textured,   flat shaded
    _vvoid FogTGRender(_CVortexTexture *pcsTexture);                       // Fogged, textured,   gouroud shaded
    // Special overloaded operators (to support aligned memory allocation for SSE)
    inline _vvoid *operator new      (_vuint32 un32Size) { return(ssemalloc(un32Size)); }
    inline _vvoid *operator new[]    (_vuint32 un32Size) { return(ssemalloc(un32Size)); }
    inline _vvoid  operator delete   (_vvoid *pvBuffer)  { return(ssefree(pvBuffer)); }
    inline _vvoid  operator delete[] (_vvoid *pvBuffer)  { return(ssefree(pvBuffer)); }
  
  // Friend classes
  friend class _CVortexPanel3DBuffer;
};

// ############################################################################################
// Inline functions implementation
// ############################################################################################
// ============================================================================================
// From exported classes
// ============================================================================================
// --------------------------------------------------------------------------------------------
// The 3D panel class
// --------------------------------------------------------------------------------------------
// ***** Constructor
_CVortexPanel3D::_CVortexPanel3D()
{
  mt_tP.un32Invisible = 0;
}

// ***** Lighting functions
_vvoid _CVortexPanel3D::_CalcIntenAmbient_(_TVortexAmbientLight *ptVAmbientLight)
{
  // ..... Use the supplied ambient light
  if(ptVAmbientLight)
  {
    mt_tP.tAmbient.un8R = ptVAmbientLight->un8R;
    mt_tP.tAmbient.un8G = ptVAmbientLight->un8G;
    mt_tP.tAmbient.un8B = ptVAmbientLight->un8B;
  }
  // ..... Use the default ambient light from the active 3D environment
  else
  {
    mt_tP.tAmbient.un8R = _pcsAct3DEnv_->mt_tDefALight.un8R;
    mt_tP.tAmbient.un8G = _pcsAct3DEnv_->mt_tDefALight.un8G;
    mt_tP.tAmbient.un8B = _pcsAct3DEnv_->mt_tDefALight.un8B;
  }
}
_vvoid _CVortexPanel3D::_FSCalcInten_(_vvoid)
{
  mt_tP.tTotal.un8R = (mt_tP.tAmbient.un8R + mt_tP.tOmni.un8R) >> 1;
  mt_tP.tTotal.un8G = (mt_tP.tAmbient.un8G + mt_tP.tOmni.un8G) >> 1;
  mt_tP.tTotal.un8B = (mt_tP.tAmbient.un8B + mt_tP.tOmni.un8B) >> 1;
}
_vvoid _CVortexPanel3D::_GSCalcInten_(_vvoid)
{
  for(_vuint8 un8Cnt = 0; un8Cnt < 4; un8Cnt++)
  {
    mt_tP.atTotal[un8Cnt].un8R = (mt_tP.tAmbient.un8R + mt_tP.atOmni[un8Cnt].un8R) >> 1;
    mt_tP.atTotal[un8Cnt].un8G = (mt_tP.tAmbient.un8G + mt_tP.atOmni[un8Cnt].un8G) >> 1;
    mt_tP.atTotal[un8Cnt].un8B = (mt_tP.tAmbient.un8B + mt_tP.atOmni[un8Cnt].un8B) >> 1;
  }
}

// ***** Clipping functions
_vvoid _CVortexPanel3D::UpdateFrame(_vvoid)
{
  if(mt_tP.un32Invisible > 0) mt_tP.un32Invisible--;
}
_vuint32 _CVortexPanel3D::Invisible(_vvoid)
{
  return(mt_tP.un32Invisible);
}
_vbool _CVortexPanel3D::IsVisible3D(_vvoid)
{
  if(_CalcBFace_()) return(_CheckExtents_());
  return(false);
}
_vbool _CVortexPanel3D::IsVisible2D(_vvoid)
{
  // ..... Temporary variables
  _vuint32 un32XMinInVis = 0, un32YMinInVis = 0, un32XMaxInVis = 0, un32YMaxInVis = 0;
  _vint32  n32AveX = 0, n32AveY = 0;
  _vuint8  un8Cnt;

  // ..... Make sure the panel has more than two points
  if(mt_tP.un8SPCount < 3)
  {
    mt_tP.un32Invisible = 4;
    return(false);
  }

  // ..... Determine the location of panel's 2D points
  for(un8Cnt = 0; un8Cnt < mt_tP.un8SPCount; un8Cnt++)
  {
    if(mt_tP.tSPoint[un8Cnt].n32X < _pcsAct3DEnv_->mt_n32MinX)
      un32XMinInVis++;
    else if(mt_tP.tSPoint[un8Cnt].n32X > _pcsAct3DEnv_->mt_n32MaxX)
      un32XMaxInVis++;

    if(mt_tP.tSPoint[un8Cnt].n32Y < _pcsAct3DEnv_->mt_n32MinY)
      un32YMinInVis++;
    else if(mt_tP.tSPoint[un8Cnt].n32Y > _pcsAct3DEnv_->mt_n32MaxY)
      un32YMaxInVis++;
    
    n32AveX += mt_tP.tSPoint[un8Cnt].n32X;
    n32AveY += mt_tP.tSPoint[un8Cnt].n32Y;
  }
  
  // ..... Checking
  if(un32XMinInVis >= mt_tP.un8SPCount)
  {
    n32AveX = (n32AveX < 0) ? -(n32AveX / mt_tP.un8SPCount) : (n32AveX / mt_tP.un8SPCount);
    mt_tP.un32Invisible = _vuint32(n32AveX) / _pcsAct3DEnv_->mt_tRB->un32Width;
    return(false);
  }
  else if(un32YMinInVis >= mt_tP.un8SPCount)
  {
    n32AveY = (n32AveY < 0) ? -(n32AveY / mt_tP.un8SPCount) : (n32AveY / mt_tP.un8SPCount);
    mt_tP.un32Invisible = _vuint32(n32AveY) / _pcsAct3DEnv_->mt_tRB->un32Height;
    return(false);
  }
  else if(un32XMaxInVis >= mt_tP.un8SPCount)
  {
    mt_tP.un32Invisible = _vuint32((n32AveX / mt_tP.un8SPCount) - _pcsAct3DEnv_->mt_n32MaxX) /
                          _pcsAct3DEnv_->mt_tRB->un32Width;
    return(false);
  }
  else if(un32YMaxInVis >= mt_tP.un8SPCount)
  {
    mt_tP.un32Invisible = _vuint32((n32AveY / mt_tP.un8SPCount) - _pcsAct3DEnv_->mt_n32MaxY) /
                          _pcsAct3DEnv_->mt_tRB->un32Height;
    return(false);
  }
  return(true);
}
_vbool _CVortexPanel3D::_CalcBFace_(_vvoid)
{
  // ..... Temporary variables
  _vfloat fDir;
  
  // ..... Calculate the back face
  fDir = mt_tP.csVPoint[0]->m_tWorld.fX * (mt_tP.tNormal.fXt - mt_tP.csVPoint[0]->m_tWorld.fX) +
         mt_tP.csVPoint[0]->m_tWorld.fY * (mt_tP.tNormal.fYt - mt_tP.csVPoint[0]->m_tWorld.fY) + 
         mt_tP.csVPoint[0]->m_tWorld.fZ * (mt_tP.tNormal.fZt - mt_tP.csVPoint[0]->m_tWorld.fZ);
  if(fDir > _vfloat(0.0))
  {
    mt_tP.un32Invisible = _vuint32(fDir / mt_tP.csVPoint[0]->Mag() * _vfloat(25.0));
    return(false);
  }
  return(true);
}
_vbool _CVortexPanel3D::_CheckExtents_(_vvoid)
{
  // ..... Temporary variables
  _vbool  bVis = false;
  _vuint8 un8Cnt;
  _vint32 n32SumZ;
  _vfloat fMinZ;
  
  // ..... Check againts MinZ
  for(un8Cnt = 0; un8Cnt < 4; un8Cnt++)
    if(mt_tP.csVPoint[un8Cnt]->m_tWorld.fZ > _pcsAct3DEnv_->mt_un32MinZ)
    {
      bVis = true;
      mt_tP.un32Invisible = 0;
      break;
    }
    
  // ..... Still visible, check againts MaxZ
  if(bVis)
  {
    fMinZ = mt_tP.csVPoint[0]->m_tWorld.fZ;
    for(un8Cnt = 0; un8Cnt < 4; un8Cnt++)
      if(mt_tP.csVPoint[un8Cnt]->m_tWorld.fZ < fMinZ) fMinZ = mt_tP.csVPoint[un8Cnt]->m_tWorld.fZ;
    
    if(fMinZ > _pcsAct3DEnv_->mt_un32MaxZ)
    {
      bVis = false;
      mt_tP.un32Invisible = _vuint32(fMinZ - _pcsAct3DEnv_->mt_un32MaxZ) >> 6;
    } 
  }
  // ..... Not visible
  else
  {
    n32SumZ = _vint32(mt_tP.csVPoint[0]->m_tWorld.fZ + mt_tP.csVPoint[1]->m_tWorld.fZ +
                      mt_tP.csVPoint[2]->m_tWorld.fZ + mt_tP.csVPoint[3]->m_tWorld.fZ);

    mt_tP.un32Invisible = _vuint32(((n32SumZ < 0) ? -n32SumZ : n32SumZ) >> 2) >> 6;
  }
  
  // ..... Return the value
  return(bVis);
}
_vbool _CVortexPanel3D::_HasVertex_(_CVortexPoint3D *pcsVPoint3D)
{
  if(mt_tP.csVPoint[0] == pcsVPoint3D) return(true);
  if(mt_tP.csVPoint[1] == pcsVPoint3D) return(true);
  if(mt_tP.csVPoint[2] == pcsVPoint3D) return(true);
  if(mt_tP.csVPoint[3] == pcsVPoint3D) return(true);
  return(false);  
}

#endif
// ============================================================================================
// End of vpanel3d.h
// ============================================================================================
