// ############################################################################################
//  test.cpp
//  The test application for VortexGE
//  -------------------------------------------------------------------------------------------
//  Copyright (C) 2004 Aloysius Indrayanto
//  This file is part of the VortexGE
//
//  This program is free software; you can redistribute it and/or modify it under the terms of 
//  the GNU Lesser General Public License as published by the Free Software Foundation; either
//  version 2.1 of the License, or (at your option) any later version.
//
//  This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
//  without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
//  See the GNU Lesser General Public License for more details.
//
//  You should have received a copy of the GNU Lesser General Public License along with this 
//  library; if not, write to the Free Software Foundation, Inc., 59 Temple Place, Suite 330, 
//  Boston, MA 02111-1307 USA.
// ############################################################################################

// ============================================================================================
// Header files
// ============================================================================================
// ***** VortexGE headers
#include <vmain.h>

// ============================================================================================
// Local functions prototype
// ============================================================================================
// ***** Initialization functions
_vbool InitializeSystem(_vvoid);
_vbool LoadImages(_vvoid);
_vbool LoadAudios(_vvoid);
_vbool CreateScene(_vvoid);

// ***** Rendering functions
_vvoid RenderPanels(_vvoid);
_vvoid RenderSprites(_vvoid);

// ============================================================================================
// Local variables declaration
// ============================================================================================
// ***** Objects
static _CVortexVideoEnv       csVidEnv;
static _CVortex3DEnv          cs3DEnv;
static _CVortexPanel3DBuffer  csWorld;
static _CVortexMatrix3D       csMat;
static _CVortexImageBuilder   csImgBuild;
static _CVortexImage          csBackImg, csTreeImg;
static _CVortexAnimatedImage  csLightImg, csArrowImg;
static _CVortexAImgControl    csArrowCtl;
static _CVortexTexture        csTex;
static _CVortexStaticSprite2D csTreeSpr, csLightSpr;
static _CVortexAudioEnv       csAudioEnv;
static _CVortexSoundBuffer    csSBBack, csSBShot;

// ***** Rendering controls
static _vbool  bGouroudShaded, bBackImg, bTransparent, bFog;
static _vuint8 un8TexType, un8ADir;

// ============================================================================================
// Main program
// ============================================================================================
int main(_vvoid)
{
  // ..... Local variables
  _vfloat           fALCol, fALColAdd;
  _TVortexColor     tInfoColor;
  _vuint64          un64Frames, un64Ticks1, un64Ticks2;
  _vfloat           fFPS = 0, fOldFPS = 1;
  _vchar            acBuff[64];
  _EVortexKeySymbol eKS;

  // ..... Initialization
  if(!InitializeSystem()) return(1);
  if(!LoadImages()) return(1);
  if(!LoadAudios()) return(1);

  // ..... Create scene
  tInfoColor = csVidEnv.CreateColor(255, 255, 255);
  csMat.Identity();
  cs3DEnv.SetDefaultOmniLightPosition(40, 200, 150);
  cs3DEnv.SetDefaultOmniLightColor(31, 31, 16);
  CreateScene();
  csWorld.SetTexture(&csTex);
  csWorld.CalcDefTexCoords();
  csWorld.SetColor(24, 31, 24);
  cs3DEnv.SetFogState(8, 16, 12, 13);
  
  // ..... Play background audio
  csSBBack.Play(true);  

  // ..... Initialize variables
  fALColAdd = 0.2;
  fALCol    = 8.0;
  bGouroudShaded = true;
  bBackImg       = true;
  bTransparent   = false;
  bFog           = true;
  un8TexType     = 2;
  un8ADir        = 0;
 
  // ..... Main loop
  VInfo(_F_, _L_, "Main loop started\n");
  csVidEnv.AutoRepeatOff();
  un64Frames = 0;
  un64Ticks1 = cs3DEnv.GetTicks();
  while(1)
  {
    // Calculate FPS
    fFPS = cs3DEnv.CalcFPS();
    if(fOldFPS != fFPS)
    {
      fOldFPS = fFPS;
      sprintf(acBuff, "%05.1f fps", fFPS);
    }
    un64Frames++;
    
    // Update everything if the application is active
    if(csVidEnv.IsActive())
    {
      // Update frame
      fALCol += fALColAdd;
      if(fALCol > 32)
      {
        fALCol = 32;
        fALColAdd = -fALColAdd;
      }
      if(fALCol < 8)
      {
        fALCol = 8;
        fALColAdd = -fALColAdd;
      }
      cs3DEnv.SetDefaultAmbientLightColor(_vuint8(fALCol), _vuint8(fALCol), _vuint8(fALCol));
      if(bGouroudShaded)
        csWorld.GSResetAmbientLight();
      else
        csWorld.FSResetAmbientLight();
      csTreeSpr.ResetAmbientLight();
      csLightImg.StepFrame(10);
      csArrowCtl.SelectNextASFrameDly(10, true);
  
      // Render scene
      if(bBackImg)
        cs3DEnv.ClearRenderBuffer(csBackImg);
      else
        cs3DEnv.ClearRenderBuffer(V_RGB(16, 8, 12));
      cs3DEnv.ClearZBuffer();
      RenderPanels();
      RenderSprites();

      // Update display
      csVidEnv.DisplayRenderImage();
      csVidEnv.SetForegroundColor(tInfoColor);
      csVidEnv.DrawText(10, 20, acBuff, 9);
      if(bBackImg)
        csVidEnv.DrawText(10, 50, "Press 'B' to change background image->solid fill", 48);
      else
        csVidEnv.DrawText(10, 50, "Press 'B' to change solid fill->background image", 48);
      if(bGouroudShaded)
        csVidEnv.DrawText(10, 65, "Press 'S' to change GS->FS", 26);
      else
        csVidEnv.DrawText(10, 65, "Press 'S' to change FS->GS", 26);
      if(un8TexType == 0)
        csVidEnv.DrawText(10, 80, "Press 'T' to use solid color", 28);
      else if(un8TexType == 1)
        csVidEnv.DrawText(10, 80, "Press 'T' to use texture", 24);
      else if(un8TexType == 2)
        csVidEnv.DrawText(10, 80, "Press 'T' to use texture without any shading", 44);
      else
      csVidEnv.DrawText(10, 80, "Press 'T' to use light color only", 33);
      if(bFog)
        csVidEnv.DrawText(10, 95, "Press 'F' to remove fog", 23);
      else
        csVidEnv.DrawText(10, 95, "Press 'F' to use fog", 20);
      if(bTransparent)
        csVidEnv.DrawText(10, 110, "Press 'N' to use solid sprite", 29);
      else
        csVidEnv.DrawText(10, 110, "Press 'N' to use transparent sprite", 35);
      csVidEnv.DrawText(10, 125, "Press 'A' to reverse the arrow direction", 40);
      csVidEnv.DrawText(10, 155, "Press 'Z' to play a one-shot sound buffer", 41);
      csVidEnv.DrawText(10, 185, "Click left mouse button to exit ...", 35);
      csVidEnv.UpdateMainWindow();
 
      // Update sound
      csAudioEnv.UpdateBuffer();
      
      // Pump event
      csVidEnv.DispatchEvent();
      if(csVidEnv.CheckEvent())
      {
        // Keyboard event
        if(csVidEnv.EventKeyPress())
          VPrint("KeyPress   : %d (%d)\n", csVidEnv.GetKeySym(), csVidEnv.GetRawKeyCode());
        else if(csVidEnv.EventKeyRelease())
        {
          eKS = csVidEnv.GetKeySym();
          if(eKS == 'b')
            bBackImg = !bBackImg;
          else if(eKS == 's')
            bGouroudShaded = !bGouroudShaded;
          else if(eKS == 't')
          {
            un8TexType++;
            if(un8TexType > 3) un8TexType = 0;
          }
          else if(eKS == 'f')
            bFog = !bFog;
          else if(eKS == 'n')
            bTransparent = !bTransparent;
          else if(eKS == 'a')
          {
            un8ADir = (un8ADir == 0) ? 1 : 0;
            csArrowCtl.SelectAnimationSet(un8ADir);
          }
          else if(eKS == 'z')
            csSBShot.Play();
          VPrint("KeyRelease : %d (%d)\n", eKS, csVidEnv.GetRawKeyCode());
        }
        // Mouse event
        else if(csVidEnv.EventButtonPress()) 
          VPrint("BtnPress   : %d\n", csVidEnv.GetBtnSym());
        else if(csVidEnv.EventButtonRelease()) 
        {
          VPrint("BtnRelease : %d\n", csVidEnv.GetBtnSym());
          if(csVidEnv.GetBtnSym() == VBS_MB1) break;
        }
      }
    }
    else
    {
      while(csVidEnv.IsInactive())
      {
        // Calculate FPS
        fFPS = cs3DEnv.CalcFPS();
        if(fOldFPS != fFPS)
        {
          fOldFPS = fFPS;
          sprintf(acBuff, "%05.1f fps", fFPS);
        }
        un64Frames++;
        
        // Update display
        if(csVidEnv.IsVisible())
        {
          csVidEnv.DisplayRenderImage();
          csVidEnv.SetForegroundColor(tInfoColor);
          csVidEnv.DrawText(10, 20, acBuff, 9);
          csVidEnv.UseDefaultBigFont();
          csVidEnv.DrawTextXCenter(150, "Inactive State", 14);
          csVidEnv.UseDefaultSmallFont();
          //csVidEnv.DrawTextXCenter(180, "Click left mouse button to exit ...", 35);
          csVidEnv.UpdateMainWindow();
        }
        
        // Pump event
        csVidEnv.DispatchEvent();
        usleep(100000);
      }
      csVidEnv.ClearEvent();
    }
  }
  un64Ticks2 = cs3DEnv.GetTicks();
  
  // ..... Deactivate audio
  csAudioEnv.DeactivateDevice();
  
  // ..... Calculate & display the avegare frame rate
  sprintf(acBuff, "Average frame rate = %.2f fps\n \nClick \"OK\" to exit ...", cs3DEnv.CalcAvgFPS(un64Frames, un64Ticks1, un64Ticks2));
  csVidEnv.MessageBox("End Program", acBuff);
  VInfo(_F_, _L_, "Program ended, average frame rate = %.2f FPS", cs3DEnv.CalcAvgFPS(un64Frames, un64Ticks1, un64Ticks2));

  // ..... OK !
  return(0);
}

// ============================================================================================
// Local functions implementation
// ============================================================================================
// --------------------------------------------------------------------------------------------
// Initialization functions
// --------------------------------------------------------------------------------------------
// ***** Initialize system
_vbool InitializeSystem(_vvoid)
{
  // ..... Temporary variables
  _vuint32 un32W = 640, un32H = 480, un32D = 2;
  _vchar   acBuff[64];

  // ..... Initialize main stuff
  if(!csVidEnv.CreateMainWindow(un32W, un32H, "Test : VortexGE X11")) return(false);
  if(!csVidEnv.CreateRenderImage(un32W, un32H)) return(false);
  if(!csVidEnv.CreateRenderBuffer(un32W / un32D, un32H / un32D)) return(false);
  if(!cs3DEnv.Initialize(&csVidEnv)) return(false);
  if(!csAudioEnv.QueryDeviceName(acBuff)) return(false);
  for(;;)
  {
    _vchar            acTmp[1024];
    _EVortexMsgRetVal eRet;
    
    if(csAudioEnv.OpenDevice(acBuff))
    {
      if(!csAudioEnv.InitializeDevice()) return(false);
      if(!csAudioEnv.ActivateDevice()) return(false);
      break;
    }
      
    sprintf(acTmp, "_CVortexAudioEnv failed to open audio device \"%s\", ", acBuff);
    strcat(acTmp, "the audio device may be in use.\n \n");
    strcat(acTmp, "Terminate all other multimedia applications or wait for about 30-60 seconds or\n");
    strcat(acTmp, "until a few minutes till any sound server (such as \"arts\") suspends itself.\n \n");
    strcat(acTmp, "Click abort to terminate this program ...\n");
    strcat(acTmp, "Click retry to try again ...\nClick ignore to continue without audio ...");
    eRet = csVidEnv.MessageBox("Failed Opening Audio Device", acTmp, EIconWarning, EBtnAbortRetryIgnore, ESizeMedium);
    if(eRet == EAbort) return(false);
    if(eRet == EIgnore)
    {
      VWarning(_F_, _L_, "Audio disabled");
      break;
    }
  }
  
  // ..... OK !
  return(true);
}

// ***** Load images
_vbool LoadImages(_vvoid)
{
  // ..... Load background image 
  if(!csImgBuild.LoadVIB("vgcachef/back2"))
  {
    if(!csImgBuild.LoadBMP("vgmediaf/bmp/back2")) return(false);
    if(!csImgBuild.SaveVIB("vgcachef/back2")) return(false); 
  }
  csImgBuild.ResizeToFitBackground();
  csImgBuild.AdjustBrightness(0.75, 0);
  if(!csImgBuild.TransferToImage(csBackImg)) return(false);
  
  // ..... Load texture image
  if(!csImgBuild.LoadVIB("vgcachef/tex2"))
  {
    if(!csImgBuild.LoadBMP("vgmediaf/bmp/tex2")) return(false);
    if(!csImgBuild.SaveVIB("vgcachef/tex2")) return(false); 
  }
  csImgBuild.Resize(128, 128);
  if(!csImgBuild.TransferToTexture(csTex)) return(false);

  // ..... Load tree sprite image
  if(!csImgBuild.LoadVIB("vgcachef/tree1"))
  {
    if(!csImgBuild.LoadBMP("vgmediaf/bmp/tree1")) return(false);
    if(!csImgBuild.SaveVIB("vgcachef/tree1")) return(false); 
  }
  csImgBuild.RoughResize(64, 0);
  if(!csImgBuild.TransferToImage(csTreeImg)) return(false);

  // ..... Load the light sprite image
  if(!csImgBuild.LoadVIB("vgcachef/light2"))
  {
    if(!csImgBuild.LoadBMP("vgmediaf/bmp/light2", true)) return(false);
    if(!csImgBuild.SaveVIB("vgcachef/light2")) return(false); 
  }
  csImgBuild.AdjustBrightness(0.5, 0, true, true, false);
  if(!csImgBuild.TransferToAnimatedImage(csLightImg, 1, 10)) return(false);
  
  // ..... Load the arrow sprite image
  if(!csImgBuild.LoadVIB("vgcachef/arrow"))
  {
    if(!csImgBuild.LoadBMP("vgmediaf/bmp/arrow", true)) return(false);
    if(!csImgBuild.SaveVIB("vgcachef/arrow")) return(false); 
  }
  if(!csImgBuild.TransferToAnimatedImage(csArrowImg, 2, 5)) return(false);
  csImgBuild.FreeImage();
  csArrowCtl.SetImageSource(&csArrowImg);
  csArrowCtl.AddAnimationSets(2);
  csArrowCtl.AddASFrame(0, 0);
  csArrowCtl.AddASFrame(0, 1);
  csArrowCtl.AddASFrame(0, 2);
  csArrowCtl.AddASFrame(0, 3);
  csArrowCtl.AddASFrame(0, 4);
  csArrowCtl.AddASFrame(1, 5);
  csArrowCtl.AddASFrame(1, 6);
  csArrowCtl.AddASFrame(1, 7);
  csArrowCtl.AddASFrame(1, 8);
  csArrowCtl.AddASFrame(1, 9);

  // ..... OK !
  return(true);
}

// ..... Load audios
_vbool LoadAudios(_vvoid)
{
  // ..... Load background audios
  if(!csSBBack.LoadVSW("vgcachef/backloop"))
  {
    //if(!csSBBack.LoadRAW("vgmediaf/raw/S16Ls11", ERF_S16LEStereo, 11025)) return(false);
    //if(!csSBBack.LoadRAW("vgmediaf/raw/S16Ls22", ERF_S16LEStereo, 22050)) return(false);
    //if(!csSBBack.LoadRAW("vgmediaf/raw/S16Ls44", ERF_S16LEStereo, 44100)) return(false);
    if(!csSBBack.LoadRAW("vgmediaf/raw/S16Ls48", ERF_S16LEStereo, 48000)) return(false);
    //if(!csSBBack.LoadRAW("vgmediaf/raw/S16Bs48", ERF_S16BEStereo, 48000)) return(false);
    //if(!csSBBack.LoadRAW("vgmediaf/raw/U16Ls48", ERF_U16LEStereo, 48000)) return(false);
    //if(!csSBBack.LoadRAW("vgmediaf/raw/U16Bs48", ERF_U16BEStereo, 48000)) return(false);
    //if(!csSBBack.LoadRAW("vgmediaf/raw/S16Lm48", ERF_S16LEMono, 48000)) return(false);
    //if(!csSBBack.LoadRAW("vgmediaf/raw/S16Bm48", ERF_S16BEMono, 48000)) return(false);
    //if(!csSBBack.LoadRAW("vgmediaf/raw/U16Lm48", ERF_U16LEMono, 48000)) return(false);
    //if(!csSBBack.LoadRAW("vgmediaf/raw/U16Bm48", ERF_U16BEMono, 48000)) return(false);
    //if(!csSBBack.LoadRAW("vgmediaf/raw/S8s48", ERF_S8Stereo, 48000)) return(false);
    //if(!csSBBack.LoadRAW("vgmediaf/raw/U8s48", ERF_U8Stereo, 48000)) return(false);
    //if(!csSBBack.LoadRAW("vgmediaf/raw/S8m48", ERF_S8Mono, 48000)) return(false);
    //if(!csSBBack.LoadRAW("vgmediaf/raw/U8m48", ERF_U8Mono, 48000)) return(false);
    if(!csSBBack.SaveVSW("vgcachef/backloop")) return(false);
  }

  // ..... Load foreground audios
  if(!csSBShot.LoadVSW("vgcachef/foreshot"))
  {
    if(!csSBShot.LoadWAV("vgmediaf/wav/S16Ls48")) return(false);
    if(!csSBShot.SaveVSW("vgcachef/foreshot")) return(false);
  }
  
  // ..... OK !
  return(true);
}

// ***** Create scene
_vbool CreateScene(_vvoid)
{
  // ..... Temporary variables
  _vfloat  afV[]   = { -60,  -50, 100,   -20,  -50, 100,   20,  -50, 100,   60,  -50, 100,
                       -60,    0, 140,   -20,    0, 140,   20,    0, 140,   60,    0, 140,
                       -60,   50, 180,   -20,   50, 180,   20,   50, 180,   60,   50, 180,
                       -60,  100, 220,   -20,  100, 220,   20,  100, 220,   60,  100, 220,
                       -60,    0, 100,   -60,   50, 140,  -60,  100, 180,  -60,  150, 220,
                        60,    0, 100,    60,   50, 140,   60,  100, 180,   60,  150, 220, 
                       -20,  150, 220,    20,  150, 220 };
  _vuint32 aun32P[] = {0,   4,  5,  1,    1,  5,  6,  2,    2,  6,  7,  3,
                       4,   8,  9,  5,    5,  9, 10,  6,    7,  6, 10, 11,
                       9,   8, 12, 13,    9, 13, 14, 10,   10, 14, 15, 11,
                       0,  16, 17,  4,    4, 17, 18,  8,    8, 18, 19, 12, 
                       3,   7, 21, 20,    7, 11, 22, 21,   11, 15, 23, 22,
                       12, 19, 24, 13,   13, 24, 25, 14,   14, 25, 23, 15 };
  _vfloat  afSPos1[] = { -30, -25, 120,   30, 75, 200 };
  _vfloat  afSPos2[] = { 140, 100, 120 };

  // ..... Create scene
  if(!csWorld.LoadVPD("vgcachef/deck"))
  {
    if(!csWorld.InitFromMemory(afV, 26, aun32P, 18)) return(false);
    csWorld.SaveVPD("vgcachef/deck");
  }
  if(!csTreeSpr.Initialize(afSPos1, 2, &csTreeImg)) return(false);
  return(csLightSpr.Initialize(afSPos2, 1, &csLightImg, ECenterCenter));
}  

// --------------------------------------------------------------------------------------------
// Rendering functions
// --------------------------------------------------------------------------------------------
// ***** Render the panels
_vvoid RenderPanels(_vvoid)
{
  if(bFog)
  {
    if(un8TexType == 0)
    {
      if(bGouroudShaded)
        csWorld.FogUGRender(csMat);
      else
        csWorld.FogUFRender(csMat);
    }
    else if(un8TexType == 1)
    {
      if(bGouroudShaded)
        csWorld.FogCGRender(csMat);
      else
        csWorld.FogCFRender(csMat);
    }
    else if(un8TexType == 2)
    {
      if(bGouroudShaded)
        csWorld.FogTGRender(csMat);
      else
        csWorld.FogTFRender(csMat);
    }
    else
    {
      csWorld.FogTNRender(csMat);
    }
  }
  else
  {
    if(un8TexType == 0)
    {
      if(bGouroudShaded)
        csWorld.UGRender(csMat);
      else
        csWorld.UFRender(csMat);
      }
    else if(un8TexType == 1)
    {
      if(bGouroudShaded)
        csWorld.CGRender(csMat);
      else
        csWorld.CFRender(csMat);
    }
    else if(un8TexType == 2)
    {
      if(bGouroudShaded)
        csWorld.TGRender(csMat);
      else
        csWorld.TFRender(csMat);
    }
    else
    {
      csWorld.TNRender(csMat);
    }
  }
}
  
// ***** Render the sprites
_vvoid RenderSprites(_vvoid)
{
  // Render the sprites
  if(bFog)
  {
    if(bTransparent)
    {
      if(un8TexType <= 2)
        csTreeSpr.FogTFRender(csMat);
      else
        csTreeSpr.FogTNRender(csMat);
    }
    else
    {
      if(un8TexType <= 2)
        csTreeSpr.FogSFRender(csMat);
      else
        csTreeSpr.FogSNRender(csMat);
    }
  }
  else
  {
    if(bTransparent)
    {
      if(un8TexType <= 2)
        csTreeSpr.TFRender(csMat);
      else
        csTreeSpr.TNRender(csMat);
    }
    else
    {
      if(un8TexType <= 2)
        csTreeSpr.SFRender(csMat);
      else
        csTreeSpr.SNRender(csMat);
    }
  }
  csLightSpr.LightRender(csMat);
  cs3DEnv.PutImage(csArrowImg, 213, 195, V_RGB(0, 0, 0));
}
  
// ============================================================================================
// End of test.cpp
// ============================================================================================
